package clarion.samples;

import clarion.system.*;
import clarion.tools.BaseSimulationTemplate;
import clarion.tools.Personality;

import java.util.ArrayList;

/**
 * This is a sample simulation running a very simple task we call the "Hello World" task. This
 * simulation is meant to demonstrate how the CLARION library can be setup. It uses mostly default
 * settings, so it should not be considered as being the definitive method for setting up a simulation
 * using the CLARION Library. It is, however, a good place to start for learning how to setup a 
 * simulation.
 * <p>
 * <u>Overview of the task:</u><br>
 * This task simulates the "Hello World" task. It uses a single agent who's job is to respond either 
 * "hello" or "goodbye" based on the salutation you give it and the salutation it has the goal to 
 * wants to give. The objective is for the agent to respond with the same salutation as the salutation it is 
 * given.
 * <p>
 * The agent has been setup using the following:
 * <ul>
 * <li>The Motivational Subsystem (MS) using the GENERIC personality and the drive strength equation 
 * (without the neural net). However, only the deficits for AFFILIATION_AND_BELONGINGNESS and 
 * AUTONOMY are specified.</li>
 * <li>The Meta Cognitive Subsystem (MCS) using the goal selecting module and the ACS level 
 * probability setting module (both using the equation, not a neural net).</li>
 * <li>The Action Centered Subsystem (ACS) with a neural network in the bottom level and RER 
 * turned on.<br>
 * The neural network uses simplified Q-learning with backpropagation and has been setup as follows:
 * </li>
 * <ul><li>Four inputs - </li>
 * <ul><li>Desired_Salutation: Hello (internally set by a goal)</li>
 * <li>Desired_Salutation: Goodbye (internally set by a goal)</li>
 * <li>Salutation: Hello (from the outside world)</li>
 * <li>Salutation: Goodbye (from the outside world)</li></ul>
 * <li>Two outputs -</li>
 * <ul><li>Hello</li>
 * <li>Goodbye</li></ul></ul>
 * </ul>
 * In addition, there are also a few optional settings that have been commented out. These options
 * are:
 * <ul>
 * <li> Turning on the benefit equation (in initializeAgentInternalSpace)</li>
 * <li> Specifying a thresholds for generalization and specialization (in initializeCLARIONAgent)</li>
 * <li> Supplying a deficit change processor (updateDeficit method which is called in run).</li>
 * </ul>
 * If you would like to see how these optional settings affect the simulation, simply remove the
 * comments in the initializeAgentInternalSpace and run methods.
 * <p>
 * It is highly recommended that you explore the documentation to get a sense of where all
 * of the various parameters are located and what options you have for configuring the system. The 
 * documentation also provides a plethora of information on how to setup pieces of the architecture.
 * If you are having difficulty figuring out how a part of the library works you can contact the 
 * CLARION library support team at:<br>
 * <a href="mailto:clarion.support@gmail.com">clarion.support@gmail.com</a>
 * <br>
 * It is advised that you check the documentation before contacting the support team as it may be
 * able to answer your question.
 * @version Configured to run with the CLARION Library, version: 6.0.5
 * @author Nick Wilson
 */
public class HelloWorld implements BaseSimulationTemplate {
	/**The sensory information.*/
	public DimensionValueCollection SensoryInputSpace;
	/**The combination of the sensory input space and the stimulus for all of the drives.*/
	public DimensionValueCollection CombinedInputSpace;
	/**The actions that can be performed in this task*/
	public ActionCollection Actions;
	/**The agent who is running this task.*/
	public CLARION Agent;
	/**The goals the agent is able to set throughout the course of the task*/
	public ArrayList <Goal> Goals;
	
	/**A counter to keep track of how many trials the agent gets correct*/
	public int CorrectCounter = 0;
	/**The number of trials to be run.*/
	public int NumberTrials = 10000;
	
	/**
	 * The main entry point for the simulation. Your task/simulation must have one of these in order
	 * to run. The CLARION Library, itself is not an executable program. It is a library for you to
	 * build your own program.
	 * @param args
	 */
	public static void main(String[] args)
	{
		//Initialize the task.
		HelloWorld hw = new HelloWorld ();
		//Initialize the sensory information space.
		hw.initializeSensoryInformationSpace();
		//Initialize the agent's internal space.
		hw.initializeAgentInternalSpace();
		//Initialize the agent.
		hw.initializeCLARIONAgent(hw.Agent);
		//Have the agent perform the task.
		hw.run(hw.Agent);
		//Report the results.
		hw.reportResults();
	}
	
	/**
	 * This method initializes the sensory input space and the combined input space.
	 */
	public void initializeSensoryInformationSpace() {
		//BEGIN initialization of the sensory input space.
		//Initialize a new dimension-value collection.
		SensoryInputSpace = new DimensionValueCollection ();
		//Add a new dimension.
		SensoryInputSpace.put("Salutations",new Dimension ("Salutations"));
		//Add new values to the dimension.
		SensoryInputSpace.get("Salutations").put("Hello", new Value("Hello"));
		SensoryInputSpace.get("Salutations").put("Goodbye", new Value("Goodbye"));
		/*Initialize the combined input space with a new dimension-value that contains the
		 * sensory input space.*/
		CombinedInputSpace = new DimensionValueCollection (SensoryInputSpace);
		//END initialization of the sensory input space.
		
		//BEGIN initialization of the drive input space.
		for(int i = 0; i < Personality.PrimaryDrives.values().length; i++)
		{
			/*For each drive in the list of drives for the personality model, add an
			 * input of type STIMULUS to the combined input space.*/
			Dimension d = new Dimension (Personality.PrimaryDrives.values()[i]);
			d.put(DriveEquation.RequiredEquationInputs.STIMULUS,
					new Value (DriveEquation.RequiredEquationInputs.STIMULUS));
			CombinedInputSpace.put(d.getID(), d);
		}
	}
	
	/**
	 * This method initializes the agent's internal space including the actions and goals, etc. 
	 */
	public void initializeAgentInternalSpace() {
		//BEGIN initialize possible actions.
		//Initialize a new action collection.
		Actions = new ActionCollection ();
		//Add a new external action to the action collection for the action "Hello".
		Actions.put("Hello", new ExternalAction ("Hello"));
		//Add a new external action to the action collection for the action "Goodbye".
		Actions.put("Goodbye", new ExternalAction ("Goodbye"));
		//END initialize possible actions.
		
		//BEGIN initialize possible goals.
		//Initialize a new collection of goals.
		Goals = new ArrayList <Goal> ();
		//Add a goal to the list of possible goals called "Greet".
		Goals.add(new Goal ("Greet"));
		//Add a dimension to the goal "Greet" called "Desired_Salutation".
		Goals.get(0).put("Desired_Salutation", new Dimension("Desired_Salutation"));
		//Add a value to the dimension "Desired_Salutation" called "Hello".
		Goals.get(0).get("Desired_Salutation").put("Hello", new Value("Hello"));
		//Add a goal to the list of possible goals called "BidFarewell".
		Goals.add(new Goal ("BidFarewell"));
		//Add a dimension to the goal "Greet" called "Desired_Salutation".
		Goals.get(1).put("Desired_Salutation", new Dimension ("Desired_Salutation"));
		//Add a value to the dimension "Desired_Salutations" called "Goodbye".
		Goals.get(1).get("Desired_Salutation").put("Goodbye", new Value("Goodbye"));
		/*Set an (optional) GLOBAL parameter in the Rule class to use the utility equation 
		 * instead of using constants.*/
		//AbstractRule.GLOBAL_UTILITY_OPTION = AbstractRule.UtilityOptions.EQUATION;
		
		//Initialize the agent providing it with the combined input space.
		Agent = new CLARION (CombinedInputSpace);
	}

	/**
	 * This method initializes the components and subsystems within the CALRION agent that are going 
	 * to be used by the agent to perform this task.
	 */
	public void initializeCLARIONAgent(CLARION Agent) {
		/*Initialize the MS. Specify the agent to which the MS will be attached, 
		 * and the personality to use for the agent (optional).*/
		MS ms = new MS (Agent, Personality.PersonalityTypes.GENERIC);
		//Initialize the MCS. Specify the agent to which the MCS will be attached.
		MCS mcs = new MCS (Agent);
		//Attach a new ACS level probability setting module to the MCS.
		mcs.attachACSLevelProbabilitySettingModule(new ACSLevelProbabilitySettingModule (ms.getBISDriveStrengths()));
		
		//BEGIN initialize relevances for "Greet" Goal.
		/*Initialize a new collection of values. This will be used to define the relevance
		 * each drive has to the "Greet" goal.*/
		ArrayList <Value> rel0 = new ArrayList <Value> ();
		/*Specify the relevance AFFILIATION_AND_BELONGINGNESS has to the "Greet" goal.*/
		rel0.add(new Value (Personality.PrimaryDrives.AFFILIATION_AND_BELONGINGNESS,1.0));
		/*Specify the relevance AUTONOMY has to the "Greet" goal.*/
		rel0.add(new Value (Personality.PrimaryDrives.AUTONOMY,0));
		/*Set the relevances for the "Greet" goal.*/
		Goals.get(0).setRelevances(rel0);
		/*Initialize a new collection of values. This will be used to define the relevance
		 * each drive has to the "BidFarewell" goal.*/
		ArrayList <Value> rel1 = new ArrayList <Value> ();
		/*Specify the relevance AFFILIATION_AND_BELONGINGNESS has to the "BidFarewell" goal.*/
		rel1.add(new Value (Personality.PrimaryDrives.AFFILIATION_AND_BELONGINGNESS,0));
		/*Specify the relevance AUTONOMY has to the "BidFarewell" goal.*/
		rel1.add(new Value (Personality.PrimaryDrives.AUTONOMY,1.0));
		/*Set the relevances for the "BidFarewell" goal.*/
		Goals.get(1).setRelevances(rel1);
		//Attach a new goal selection module to the MCS.
		mcs.attachGoalSelectionModule(new GoalSelectionModule (ms.getAllDriveStrengths(), Goals));
		//Initialize the Goal Structure. Specify the agent to which the GS will be attached. 
		GoalStructure gs = new GoalStructure (Agent);
		//Add the list of possible goals to the goal structure.
		gs.addPossibleGoals(Goals);
		//Initialize the ACS. Specify the agent to which the ACS will be attached.
		ACS acs = new ACS (Agent);
		
		/*Set an (optional) local parameter in the ACS that was just initialized for task specific
		 * specialization and generalization thresholds.*/
		acs.Refiner.SPECIALIZATION_THRESHOLD1 = - 1;
		acs.Refiner.GENERALIZATION_THRESHOLD1 = -.1;
		acs.Refiner.IG_OPTION = RuleRefiner.IG_OPTIONS.PERFECT;
		
		/*Initialize a new dimension-value collection to setup the input layer of the neural net for 
		 * the bottom level of the ACS. Specify the sensory input space.*/
		DimensionValueCollection netInput = new DimensionValueCollection (SensoryInputSpace);
		/*Add the goals to the net input collection so the neural net can handle the goals 
		 * when they are set.*/
		for(Goal g : Goals)
			netInput.putAll(g);
		/*Initialize a simplified Q-learning backpropagation neural net to be used in the bottom 
		 * level of the ACS.*/
		SimplifiedQBPNet acsbp = new SimplifiedQBPNet(netInput.values(),2,Actions);
		//Add the neural net to the bottom level of the ACS.
		acs.addBLModule(acsbp);
	}
	
	/**
	 * This method runs the agent through the Hello World task.
	 */
	public void run(CLARION Agent) {
		//For keeping track of the feedback.
		int feedback = 0;
		//For keeping track of the chosen action.
		AbstractAction ChosenAction;
		//Run the task for the specified number of trials.
		for(int i = 0; i < NumberTrials; i++)
		{
			//Randomly choose a salutation to give.
			if(Math.random() > .5)
			{
				//Specify the salutation to give as "Hello".
				CombinedInputSpace.get("Salutations").get("Hello").setActivation(1);
				CombinedInputSpace.get("Salutations").get("Goodbye").setActivation(0);
				//Specify the stimulus for the AFFILIATON_AND BELONGINNESS drive.
				CombinedInputSpace.get(Personality.PrimaryDrives.AFFILIATION_AND_BELONGINGNESS)
					.get(DriveEquation.RequiredEquationInputs.STIMULUS).setActivation(1);
				//Specify the stimulus for the AUTONOMY drive.
				CombinedInputSpace.get(Personality.PrimaryDrives.AUTONOMY)
					.get(DriveEquation.RequiredEquationInputs.STIMULUS).setActivation(0);
				System.out.print("You say, \"Hello\". ");
			}
			else
			{
				//Specify the salutation to give as "Goodbye".
				CombinedInputSpace.get("Salutations").get("Hello").setActivation(0);
				CombinedInputSpace.get("Salutations").get("Goodbye").setActivation(1);
				//Specify the stimulus for the AFFILIATON_AND BELONGINNESS drive.
				CombinedInputSpace.get(Personality.PrimaryDrives.AFFILIATION_AND_BELONGINGNESS)
					.get(DriveEquation.RequiredEquationInputs.STIMULUS).setActivation(0);
				//Specify the stimulus for the AUTONOMY drive.
				CombinedInputSpace.get(Personality.PrimaryDrives.AUTONOMY)
					.get(DriveEquation.RequiredEquationInputs.STIMULUS).setActivation(1);
				System.out.print("You say, \"Goodbye\". ");
			}
			if(i == 0)
			{
				/*Have the agent perceive the world. We are trial 0, so no feedback has been 
				 * provided yet.*/
				Agent.perceive(CombinedInputSpace);
			}
			else
			{
				/*Have the agent perceive the world including the feedback for the last 
				action it performed.*/
				Agent.perceive(CombinedInputSpace, feedback);
			}
			//Have the agent choose and action.
			ChosenAction = Agent.act();
			if(ChosenAction.equals(Actions.get("Hello")))
			{
				//The agent said "Hello".
				System.out.print("The agent said, \"Hello\". ");
				if(CombinedInputSpace.get("Salutations").get("Hello").getActivation() == 1.0)
				{
					//The agent was right.
					System.out.println("The agent was correct.");
					//Record the agent's success.
					CorrectCounter++;
					//Give positive feedback.
					feedback = 1;
				}
				else
				{
					//The agent was wrong.
					System.out.println("The agent was incorrect.");
					//Give negative feedback.
					feedback = 0;
				}
			}
			if(ChosenAction.equals(Actions.get("Goodbye")))
			{
				//The agent said "Goodbye".
				System.out.print("The agent said, \"Goodbye\". ");
				if(CombinedInputSpace.get("Salutations").get("Goodbye").getActivation() == 1.0)
				{
					//The agent was right.
					System.out.println("The agent was correct.");
					//Record the agent's success.
					CorrectCounter++;
					//Give positive feedback.
					feedback = 1;
				}
				else
				{
					//The agent was wrong.
					System.out.println("The agent was incorrect.");
					//Give negative feedback.
					feedback = 0;
				}
			}
			/*Optionally, call the deficit change processor to update the deficits for the agent
			 * based on the chosen action.*/
			updateDeficit(Agent, ChosenAction);
		}
	}
	
	/**
	 * This method reports the results of the simulation.
	 */
	public void reportResults() {
		System.out.println("The agent gave the correct salutation " + CorrectCounter + " times.");
		System.out.println("The agent learned the following rules through RER:");
		for(AbstractRule r : Agent.getACS().getRERRules())
		{
			//Report the rules the agent learned using RER.
			System.out.println(r.toString());
		}
	}
	
	/**
	 * This is an optional method for adjusting the deficit based on task specific information.
	 * This method is known as the "Deficit Change Processor".
	 * @param Agent The agent whose deficit you wish to change.
	 * @param ChosenAction The chosen action.
	 */
	public void updateDeficit (CLARION Agent, AbstractAction ChosenAction)
	{
		Drive ab  = Agent.getMS().getDrive(Personality.PrimaryDrives.AFFILIATION_AND_BELONGINGNESS);
		Drive a  = Agent.getMS().getDrive(Personality.PrimaryDrives.AUTONOMY);
		if(ChosenAction.getID().equals("Hello"))
		{
			/*If the agent said "Hello", decrease the rate of change for the 
			 * AFFILIATON_AND_BELONGINGNESS drive. Increase the rate of change for the 
			 * AUTONOMY drive.*/
			ab.setDeficitChangeRate(.99);
			a.setDeficitChangeRate(1.01);
			/*Adjust the deficit so it cannot be greater than 1. This is not a requirement, but is
			 * mainly intended to demonstrate that the drive, itself, can also be changed
			 * in the deficit change processor.*/
			if(a.getDeficit() > 1)
				a.setDeficit(1);
		}
		else 
		{
			if(ChosenAction.getID().equals("Goodbye"))
			{
				/*If the agent said "Goodbye", decrease the rate of change for the 
				 * AUTONOMY drive. Increase the rate of change for the 
				 * AFFILIATON_AND_BELONGINGNESS drive.*/
				a.setDeficitChangeRate(.99);
				ab.setDeficitChangeRate(1.01);
				/*Adjust the deficit so it cannot be greater than 1.*/
				if(ab.getDeficit() > 1)
					ab.setDeficit(1);
			}
		}
		
	}
}
