package clarion.system;

import java.util.Collection;

/**
 * This class implements an ACS level probability setting module within CLARION.
 * <p>
 * <b>Usage:</b>
 * <p>
 * This module can be optionally added to the MCS if the user wishes to have the MCS provide a probability
 * measure to the ACS to influence level selection. This module uses an implicit module to determine the
 * probability measure. The user can decide, during initialization, the implicit module they wish to use.
 * <p>
 * Note that the level probability setting implicit module MUST have an output chunk that has an ID that 
 * specifies which level probabilities are to be set. The ID must be either "TL" (for top level probability) or 
 * "BL" (for bottom level probability).
 * <p>
 * If this module is initialized without specifying an implicit module to use, the ACSLevelProbabilitySettingEquation 
 * will be used.
 * @version 6.0.4
 * @author Nick Wilson
 */
public class ACSLevelProbabilitySettingModule {
	
	/**The implicit module used for level probability setting.*/
	private AbstractImplicitModule LevelProbabilitySettingImplicitModule;
	
	/**
	 * Initializes the ACS level probability setting module using the ACS level probability setting equation.
	 * @param AffiliatedDrives The drives affiliated with ACS level selection probability setting.
	 */
	public ACSLevelProbabilitySettingModule (Collection <DriveStrength> AffiliatedDrives)
	{
		DimensionlessOutputChunkCollection occ = new DimensionlessOutputChunkCollection ();
		occ.put("TL", new DimensionlessOutputChunk("TL"));
		occ.put("BL", new DimensionlessOutputChunk("BL"));
		DriveStrengthCollection dsc = new DriveStrengthCollection (AffiliatedDrives);
		LevelProbabilitySettingImplicitModule = new ACSLevelProbabilitySettingEquation (dsc.toDimensionValueCollection().values(), occ);
	}
	
	/**
	 * Initializes the ACS level probability setting module given the implicit module specified.
	 * <p>
	 * The input nodes of the specified implicit module must be specified as value 
	 * objects that have the same ID as the drives in the MS that you wish to use for 
	 * setting ACS level probabilities.
	 * <p>
	 * Note that the specified level probability setting implicit module MUST have an output chunk that has an ID that 
	 * specifies which level probabilities are to be set. The ID must be either "TL" (for top level probability) or 
	 * "BL" (for bottom level probability). If no such output chunk(s) exist(s), this constructor throws an exception.
     * If the specified implicit module contains more than the aforementioned output chunks in its output, those other 
     * chunks will be ignored.
	 * @param im The level probability setting implicit module to use.
	 * @throws InvalidFormatException If the specified implicit module does not 
	 * contain an output chunk with a dimension that has an ID "TL\" and/or "BL".
	 */
	public ACSLevelProbabilitySettingModule (AbstractImplicitModule im) throws InvalidFormatException
	{
		boolean check = false;
		for(AbstractOutputChunk o : im.Output.values())
		{
			if(o.getID().equals("TL") || o.getID().equals("BL"))
			{
				check = true;
				break;
			}
		}
		if(!check)
			throw new InvalidFormatException ("The specified implicit module does not contain an " +
			"output chunk with a dimension that has an ID \"TL\" and/or \"BL\".");
		LevelProbabilitySettingImplicitModule = im;
	}
	
	/**
	 * Gets the probability of using the top level of the ACS given the drive strengths
	 * specified.
	 * @param DriveStrengths The drive strengths to use for calculating the probability.
	 * @return The probability of using the top level.
	 */
	public double getTopLevelProbability (DriveStrengthCollection DriveStrengths)
	{
		LevelProbabilitySettingImplicitModule.setInput(DriveStrengths.toDimensionValueCollection());
		LevelProbabilitySettingImplicitModule.forwardPass();
		return LevelProbabilitySettingImplicitModule.Output.get("TL").getActivation();
	}
	
	/**
	 * Gets the probability of using the bottom level of the ACS given the drive strengths
	 * specified.
	 * @param DriveStrengths The drive strengths to use for calculating the probability.
	 * @return The probability of using the bottom level.
	 */
	public double getBottomLevelProbability (DriveStrengthCollection DriveStrengths)
	{
		LevelProbabilitySettingImplicitModule.setInput(DriveStrengths.toDimensionValueCollection());
		LevelProbabilitySettingImplicitModule.forwardPass();
		return LevelProbabilitySettingImplicitModule.Output.get("BL").getActivation();
	}
	
	/**
	 * Gets the implicit module used by this module. This method is primarily meant for
	 * internal purposes. It should NOT be used outside of the CLARION Library for purposes 
	 * other than reporting the internal state.
	 * @return The implicit module used by this module.
	 */
	public AbstractImplicitModule getImplicitModule ()
	{
		return LevelProbabilitySettingImplicitModule;
	}
}
