package clarion.system;

import java.util.Collection;
import java.util.Map;

/**
 * This class implements a chunk within CLARION. It extends the DimensionValueCollection class.
 * This class is abstract and therefore cannot be instantiated on its own.
 * <p>
 * <b>Usage:</b>
 * <p>
 * In it's most basic form, a chunk is essentially just a dimension-value collection that contains an ID. It is
 * a representational object used most often by the top level of the various CLARION subsystems. In addition, some
 * chunks may be linked to the output side of various explicit and implicit modules and can be used to pass pertinent information
 * and instructions between the subsystems.
 * <p>
 * It is required that you specify a way of identifying the chunk (eg. name, number, etc) when 
 * you initialize this class. Once this ID is set, it cannot be changed. This ID is used as the 
 * key for this chunk when it is placed inside a map.
 * <p>
 * <b>Known Subclasses:</b><br>
 * <ul>
 * <li>AbstractAction</li>
 * <li>AbstractOutputChunk</li>
 * <li>DimensionlessOutputChunk</li>
 * <li>DriveStrength</li>
 * <li>ExternalAction</li>
 * <li>Goal</li>
 * <li>GoalAction</li>
 * <li>WorkingMemoryAction</li>
 * </ul>
 * @version 6.0.4
 * @author Nick Wilson
 */
public abstract class AbstractChunk extends DimensionValueCollection {
	private static final long serialVersionUID = -1327318705390162332L;

	/**Identifies the chunk. A name (in the form of a string) is simply one option for identifying a
	 * chunk.*/
	private Object ID;	
	
	/**
	 * Initializes the chunk with the ID specified.
	 * @param id The ID of the chunk.
	 */
	public AbstractChunk (Object id)
	{
		ID = id;
	}
	
	/**
	 * Initializes the chunk with the specified ID and dimensions.
	 * @param id The ID of the chunk.
	 * @param dims The dimensions for the chunk.
	 */
	public AbstractChunk (Object id, Collection <? extends Dimension> dims)
	{
		super(dims);
		ID = id;
	}
	
	/**
	 * Initializes the chunk with the specified ID and map of dimensions.
	 * @param id The ID of the chunk.
	 * @param dims The map of dimensions for the chunk.
	 */
	public AbstractChunk (Object id, Map <? extends Object, ? extends Dimension> dims)
	{
		super(dims);
		ID = id;
	}
	
	/**
	 * Gets the ID of this chunk.
	 * @return The ID of this chunk.
	 */
	public Object getID ()
	{
		return ID;
	}
	
	/**
	 * Checks to see if the specified chunk has the same ID as this chunk.
	 * @param chunk The chunk whose ID you want to compare to this chunk's ID.
	 * @return True if the specified chunk has the same ID as this chunk, otherwise false.
	 */
	public boolean equalsID (Object chunk)
	{
		if(chunk == this)
			return true;
		if(!(chunk instanceof AbstractChunk))
			return false;
		if(ID == null && ((AbstractChunk)chunk).ID == null)
			return true;
		if((((AbstractChunk)chunk).ID == null && ID != null) || (ID == null && ((AbstractChunk)chunk).ID != null))
			return false;
		return ((AbstractChunk)chunk).ID.equals(ID);
	}
	
	/**
	 * Checks to see if the specified object is a chunk and if the ID of 
	 * the specified chunk is the same as this chunk. It also checks to make sure
	 * all of the dimensions within the chunk are equal.
	 * @param chunk The object to compare to this chunk.
	 * @return True if the two chunks are equal, otherwise false.
	 */
	public boolean equals (Object chunk)
	{
		if(chunk == this)
			return true;
		if(!(chunk instanceof AbstractChunk))
			return false;
		if(!equalsID(chunk))
			return false;
		for (Map.Entry<? extends Object, ? extends Dimension> e : ((AbstractChunk)chunk).entrySet())
			if(!containsKey(e.getKey()) || !containsValue(e.getValue()))
				return false;
		return true;
	}
	
	/**
	 * Clones the chunk (including all of it's dimensions).
	 * @return A copy of the chunk.
	 */
	public abstract AbstractChunk clone ();
	
	public String toString()
	{
		if(!(ID == null) && !(ID instanceof AbstractChunk))
			return "Chunk ID - " + ID.toString() + ":\n" + super.toString();
		return "Chunk ID - ?:\n" + super.toString();
	}
}
