package clarion.system;

import java.util.LinkedList;

/**
 * This class implements an explicit module within CLARION.
 * This class is abstract and therefore cannot be instantiated on its own.
 * <p>
 * <b>Usage:</b>
 * <p>
 * Any class that extends this class can be used on the top level of the CLARION subsystems. This class
 * mainly provides a framework for building explicit modules for the top level.
 * <p>
 * <b>Known Subclasses:</b><br>
 * <ul>
 * <li>AbstractFixedRule</li>
 * <li>AbstractIRLRule</li>
 * <li>AbstractRule</li>
 * <li>RefineableRule</li>
 * </ul>
 * <p>
 * This class contains both global (static) and local constants. The default is to use the local 
 * constants. If you want to change any of the global constants, you need to do so before any
 * instances of this class are initialized.
 * @version 6.0.4
 * @author Nick Wilson
 */
public abstract class AbstractExplicitModule{
	
	/**The initalBLA. Used to calculate BLA*/
	public static double GLOBAL_INITIAL_BLA = 0;
	/**The initalBLA. Used to calculate BLA*/
	public double INITIAL_BLA = GLOBAL_INITIAL_BLA;
	/**Time stamps for each usage. Used for calculating BLA.*/
	private LinkedList <Long> T = new LinkedList <Long> ();
	/**Constant c. Used for calculating BLA*/
	public static double GLOBAL_C = 2;
	/**Constant d. Used for calculating BLA*/
	public static double GLOBAL_D = 0.5;
	/**The weight. Used to calculate support.*/
	public static double GLOBAL_WEIGHT = 1;
	/**Constant c. Used for calculating BLA*/
	public double C = GLOBAL_C;
	/**Constant d. Used for calculating BLA*/
	public double D = GLOBAL_D;
	/**The weight. Used to calculate support.*/
	public double WEIGHT = GLOBAL_WEIGHT;
	
	/**Perception time. Used for calculating response time.*/
	public static double GLOBAL_PERCEPTION_TIME = .200;
	/**Perception time. Used for calculating response time.*/
	public double PERCEPTION_TIME = GLOBAL_PERCEPTION_TIME;
	/**Decision time. Used for calculating response time.*/
	public static double GLOBAL_DECISION_TIME = .350;
	/**Decision time. Used for calculating response time.*/
	public double DECISION_TIME = GLOBAL_DECISION_TIME;
	/**Actuation time. Used for calculating response time.*/
	public static double GLOBAL_ACTUATION_TIME = .500;
	/**Actuation time. Used for calculating response time.*/
	public double ACTUATION_TIME = GLOBAL_ACTUATION_TIME;
	
	/**
	 * Gets the BLA.
	 * @return The BLA.
	 */
	public double getBLA()
	{
		double BLA = 0;
		for(long i : T)
		{
			BLA += Math.pow(i,-D);
		}
		BLA *= C;
		return BLA + INITIAL_BLA;
	}
	
	/**
	 * Adds a time stamp to this rule. The time stamp is used for calculating BLA.
	 * @param stamp The time stamp to add.
	 */
	public void addTimeStamp (long stamp)
	{
		T.add(stamp);
	}
	
	/**
	 * Gets the response time.
	 * @return The response time.
	 */
	public double getResponseTime ()
    {
    	return PERCEPTION_TIME + DECISION_TIME + ACTUATION_TIME;
    }
	
	public abstract double getUtility();
	
	public abstract double getSupport(DimensionValueCollection CurrentInput);
	
	public abstract boolean equals (Object R);
}
