package clarion.system;

/**
 * This class implements a fixed rule within CLARION. It extends the AbstractRule class.
 * This class is abstract and therefore cannot be instantiated on its own. All user defined 
 * fixed rules MUST extend this class.
 * <p>
 * <b>Usage:</b>
 * <p>
 * Fixed rules are rules within CLARION that are more complex than a simple Condition ---> Action pairing.
 * Some examples of fixed rules include:<br>
 * <ul>
 * <li>If two dimension-value pairs within the state are activated, then set the value of the action
 * chunk to the product of the two activated dimension-value pairs in the condition.</li>
 * <li>Set the value of the action chunk to the sum of all the activated values in a given dimension of 
 * the current sensory input.</li>
 * <li>If a given dimension is activated, then activate a particular dimension-value pair in
 * the action chunk.</li>
 * <li>etc.</li>
 * </ul>
 * <p>
 * Fixed rules must contain an action chunk (although a condition chunk is optional). The values of the 
 * dimension-value pairs within an action chunk can be manipulated by the rule (although the structure
 * may not be changed).
 * <p>
 * All operations performed by the fixed rule (including matching the condition) must be specified 
 * within the "getSupport" method. The getSupport method returns a value which is equal to the support
 * for the rule. Fixed rules are no different than any other type of rule in terms of selection. They
 * will compete with the other rules for selection based on their support and utility.
 * <p>
 * Note also that as an option you can also override the getUtility function. By default the getUtility function
 * for a fixed rule will simply return the DEFAULT_UTILITY constant.
 * <p>
 * This class contains both global (static) and local constants. The default is to use the local 
 * constants. If you want to change any of the global constants, you need to do so before any
 * instances of this class are initialized.
 * @version 6.0.4
 * @author Nick Wilson
 */
public abstract class AbstractFixedRule extends AbstractRule{
	
	/**The default utility for the fixed rule.*/
	public static double GLOBAL_DEFAULT_UTILITY = 1;
	/**The default utility for the fixed rule.*/
	public double DEFAULT_UTILITY = GLOBAL_DEFAULT_UTILITY;
	
	
	/**
	 * Initializes the fixed rule with the specified action.
	 * @param act The action for the new rule.
	 */
	public AbstractFixedRule (AbstractAction act)
	{
		super(new GeneralizedConditionChunk(), act);
	}
	
	/**
	 * Initializes the fixed rule with the specified condition and action. This constructor
	 * is used in cases where you want to create a fixed rule with a standard condition chunk.
	 * @param cond The condition for the new rule.
	 * @param act The action for the new rule.
	 */
	public AbstractFixedRule (GeneralizedConditionChunk cond, AbstractAction act)
	{
		super(cond, act);
	}
	
	/**
	 * Performs the user specified operations for the fixed rule. This method is written by the 
	 * user to handle matching of the current input to the rule as well other operations as 
	 * deemed necessary.
	 * @param CurrentInput The current input.
	 * @return The support for the fixed rule.
	 */
	public abstract double getSupport (DimensionValueCollection CurrentInput);
	
	/**
	 * Gets the utility for this fixed rule. By default the utility of a fixed rule is always set to the
	 * default utility constant. However the user has the option of defining their own utility function by 
	 * simply overriding this method.
	 * @return The utility of the fixed rule.
	 */
	public double getUtility()
	{
		return DEFAULT_UTILITY;
	}
	
	/**
	 * This method is overridden by the user to handle condition matching when comparing
	 * two rules. The general rule of thumb is to have this method simply return false.
	 * @param R The object to compare to this rule.
	 * @return True if the specified object is equal to this rule, otherwise false.
	 */
	public abstract boolean equals (Object R);
	
	/**
	 * This method is overridden by the user to handle the comparing of two
	 * two rules to see if one covers the other. The general rule of thumb is to have this 
	 * method simply return false.
	 * @param R the object to compare to this rule.
	 * @return True if this rule covers the specified object, otherwise false.
	 */
	public abstract boolean covers (Object R);
}
