package clarion.system;

/**
 * This class implements a drive within CLARION.
 * <p>
 * <b>Usage:</b>
 * <p>
 * This class represents a drives, which exist at the bottom-level of the MS within the
 * behavioral approach and/or behavioral inhibition systems (or a special "neither" behavioral system).
 * <p>
 * The drive class uses an implicit module in order to obtain a drive strength. The implicit module
 * can be user defined, however if it is not, then the drive will use the DriveEquation as its default.
 * <p>
 * If you are using the drive equation, the drive input MUST contain a dimension with an ID 
 * equal to the ID of this drive that contains a value with the ID: <b>STIMULUS</b> 
 * (of enumerated type <b>RequiredEquationInputs</b>) and a value with the ID: <b>SYSTEM_GAIN</b> 
 * (of enumerated type <b>RequiredEquationInputs</b>).
 * <p>
 * All other dimensions and values in the drive input will be ignored if you 
 * are using the equation.
 * <p>
 * Note that if the drive belongs to both the BIS and BAS, the SYSTEM_GAIN should be set
 * to the average of the gains of the two drives.
 * <p>
 * The formula for the drive equation is as follows:
 * <br>
 * <b>DriveStrength = (DRIVE_GAIN * SYSTEM_GAIN * UNIVERAL_GAIN) * Deficit * Stimulus + Baseline</b>
 * <p>
 * If a neural network is used as the implicit module for the drive, it should be pre-trained before it 
 * is used as drives are not (generally) runtime trainable (although the you can certainly use a runtime 
 * trainable implicit module if you provide your own method for training it during runtime).
 * <p>
 * Note that if a user defined implicit module is used, it MUST have an output chunk on the 
 * output layer that has an ID that is equal to the ID of the drive for which it is being used and 
 * that output chunk MUST be of type DriveStrength. If no such output chunk exists, the drive will fail to initialize.
 * <p>
 * If the implicit module contains more than one output chunk in the output layer, all other output chunks besides the 
 * drive strength chunk for this drive will be ignored.
 * <p>
 * However, it should also be noted, that a user COULD setup a single implicit module that has ALL of the drives linked 
 * to a drive strength on the output layer of that single implicit module and then use that implicit module to obtain
 * drive strengths for ALL of the drives in the MS. Using a single implicit module COULD reduce the amount of time it takes 
 * to pre-train the drives in the MS.
 * <p>
 * If you are using a user defined implicit module and if that module is not using any of the parameters specified in
 * the drive equation they will just be ignored.
 * <p>
 * This class contains both global (static) and local constants. The default is to use the local 
 * constants. If you want to change any of the global constants, you need to do so before any
 * instances of this class are initialized.
 * @version 6.0.4
 * @author Nick Wilson
 */
public class Drive {
	
	/**Identifies the drive. A name (in the form of a string) is simply one option for identifying a
	 * drive.*/
	private Object ID;
	
	/**The hash for this object. When this item is cloned, so is this hash.*/
	private int hash;
	
	/**The implicit module that is used to get drive strengths.*/
	private AbstractImplicitModule DriveMod;

	/**The drive strength associated with this drive.*/
	protected DriveStrength DS;
	
	/**The universal gain for all drives*/
	public static double GLOBAL_UNIVERSAL_GAIN = 1;
	/**The rate of change for the deficit.*/
	public static double GLOBAL_DEFICIT_CHANGE_RATE = 1;
	/**The universal gain for all drives*/
	public double UNIVERSAL_GAIN = GLOBAL_UNIVERSAL_GAIN;
	/**The rate of change for the deficit.*/
	private double DEFICIT_CHANGE_RATE = GLOBAL_DEFICIT_CHANGE_RATE;
	
	/**The drive deficit represented as a value object.*/
	private Value Deficit;
	/**The initial (pre-decay) drive deficit. This is used to calculate the baseline when a deficit
	 * portion is being used.*/
	private Value InitialDeficit;
	/**The gain for the particular drive*/
	public static double GLOBAL_GAIN = 1;
	/**The baseline for this particular drive*/
	public static double GLOBAL_BASELINE = .01;
	/**The gain for the particular drive*/
	public double GAIN = GLOBAL_GAIN;
	/**The baseline for this particular drive*/
	public double BASELINE = GLOBAL_BASELINE;
	
	/**
	 * Initializes the drive to use the drive equation for reporting drive strengths. This is
	 * the default option if an implicit module is not specified. This constructor 
	 * also initializes the deficit to the specified deficit and sets the ID to the 
	 * specified ID. Once the drive has been initialized, the initial deficit cannot be 
	 * changed.
	 * @param id The ID to set for this drive.
	 * @param deficit The initial drive deficit.
	 */
	public Drive (Object id, double deficit)
	{
		InitialDeficit = new Value (this, deficit);
		Deficit = InitialDeficit.clone();
		ID = id;
		DimensionValueCollection Input = new DimensionValueCollection();
		Dimension d = new Dimension(ID);
		Input.put(d.getID(), d);
		d.put(DriveEquation.RequiredEquationInputs.DRIVE_GAIN, 
				new Value (DriveEquation.RequiredEquationInputs.DRIVE_GAIN));
		d.put(DriveEquation.RequiredEquationInputs.SYSTEM_GAIN, 
				new Value (DriveEquation.RequiredEquationInputs.SYSTEM_GAIN));
		d.put(DriveEquation.RequiredEquationInputs.UNIVERSAL_GAIN, 
				new Value (DriveEquation.RequiredEquationInputs.UNIVERSAL_GAIN));
		d.put(DriveEquation.RequiredEquationInputs.STIMULUS, 
				new Value (DriveEquation.RequiredEquationInputs.STIMULUS));
		d.put(DriveEquation.RequiredEquationInputs.DEFICIT, 
				new Value (DriveEquation.RequiredEquationInputs.DEFICIT));
		d.put(DriveEquation.RequiredEquationInputs.BASELINE, 
				new Value (DriveEquation.RequiredEquationInputs.BASELINE));
		DriveStrengthCollection dsc = new DriveStrengthCollection();
		dsc.put(ID, new DriveStrength(ID));
		DriveMod = new DriveEquation(ID,Input.values(),dsc);
		DS = (DriveStrength)DriveMod.Output.get(ID);
		hash = System.identityHashCode(this);
	}
	
	/**
	 * Initializes the drive to use the specified implicit module for reporting 
	 * drive strengths. Also initializes the deficit to the specified deficit and sets the 
	 * ID to the specified ID. Once the drive has been initialized, the initial deficit 
	 * cannot be changed.
	 * <p>
	 * Note that the specified implicit module MUST have an output chunk on the 
	 * output layer that has an ID that is equal to the ID of this drive instance and 
	 * that output chunk MUST be of type DriveStrength. If no such output chunk exists, the 
	 * drive will fail to initialize.
     * <p>
     * If the implicit module contains more than one output chunk in the output 
     * layer, all other output chunks besides the drive strength chunk for this drive 
     * will be ignored.
     * <p>
     * However, it should be noted, that a user COULD setup a single implicit module that has 
     * ALL of the drives linked to a drive strength on the output layer of that single implicit 
     * module and then use that implicit module to obtain drive strengths for ALL of the drives 
     * in the MS. Using a single implicit module COULD reduce the amount of time it takes to 
     * pre-train the drives in the MS.
	 * @param id The ID to set for this drive.
	 * @param deficit The initial drive deficit.
	 * @param im The implicit module to use for this drive.
	 * @throws InvalidFormatException If the specified implicit module does not contain an output 
	 * chunk with an ID equal to the ID of this drive or the output chunk has the correct ID but 
	 * is not of type DriveStrength.
	 */
	public Drive (Object id, double deficit, AbstractImplicitModule im) throws InvalidFormatException
	{
		InitialDeficit = new Value (this, deficit);
		Deficit = InitialDeficit.clone();
		ID = id;
		DS = null;
		for(AbstractOutputChunk o : im.Output.values())
		{
			if(o.getID().equals(ID) && (o instanceof DriveStrength))
			{
				DS = (DriveStrength)o;
				break;
			}
		}
		if(DS == null)
			throw new InvalidFormatException ("The specified implicit module does " +
					"not contain an output chunk with an ID equal to the ID of this drive or " +
					"the output chunk has the correct ID but is not of type DriveStrength.");
		DriveMod = im;
		hash = System.identityHashCode(this);
	}
    
	/**
	 * Calculates the drive strength given the specified drive input. In this method, the drive 
	 * specific parameters are added to the drive input before setting it as the input for 
	 * drive's implicit module. 
	 * <p>
	 * If you are using a user defined implicit module and if that module is not using any of the parameters 
	 * specified in the drive equation they will just be ignored.
	 * <p>
	 * If you are using the drive equation, the drive input MUST contain a dimension with an ID 
	 * equal to the ID of this drive that contains a value with the ID: <b>STIMULUS</b> 
	 * (of enumerated type <b>RequiredEquationInputs</b>) and a value with the ID: <b>SYSTEM_GAIN</b> 
	 * (of enumerated type <b>RequiredEquationInputs</b>).
	 * <p>
	 * All other dimensions and values in the drive input will be ignored if you 
	 * are using the equation.
	 * <p>
	 * If the drive input does not contain the aforementioned dimension and value and you 
	 * are using the equation, this method will throw an exception.
	 * <p>
	 * Note that if your drive belongs to both the BIS and BAS, the SYSTEM_GAIN should be set
	 * to the average of the gains of the two drives.
	 * @param DriveInput The input for the drive.
	 * @return The strength of the drive.
	 * @throws InvalidFormatException If the drive is using the equation and the 
	 * drive input does not contain a dimension with an ID equal to the ID of this drive 
	 * and a value within that dimension that has the ID: STIMULUS (of enumerated type RequiredEquationInputs) and a value 
	 * with the ID: SYSTEM_GAIN (of enumerated type RequiredEquationInputs).
	 */
	public DriveStrength calculateDriveStrength (DimensionValueCollection DriveInput) throws InvalidFormatException
	{
		Dimension d = null;
		if(DriveInput != null)
			d = DriveInput.get(ID);
		if(d == null)
			throw new InvalidFormatException ("The drive input does not contain a dimension " +
					"with the same ID as this drive.");
		if(DriveMod instanceof DriveEquation)
		{
			d.put(DriveEquation.RequiredEquationInputs.UNIVERSAL_GAIN, 
					new Value (DriveEquation.RequiredEquationInputs.UNIVERSAL_GAIN,UNIVERSAL_GAIN));
			d.put(DriveEquation.RequiredEquationInputs.DRIVE_GAIN, 
					new Value (DriveEquation.RequiredEquationInputs.DRIVE_GAIN,GAIN));
			d.put(DriveEquation.RequiredEquationInputs.DEFICIT, 
					new Value (DriveEquation.RequiredEquationInputs.DEFICIT,Deficit.getActivation()));
			d.put(DriveEquation.RequiredEquationInputs.BASELINE, 
					new Value (DriveEquation.RequiredEquationInputs.BASELINE,BASELINE));
		}
		DriveMod.setInput(DriveInput);
		DriveMod.forwardPass();
		return DS;
	}
	
	/**
	 * Gets the ID of this drive.
	 * @return The ID of this drive.
	 */
	public Object getID ()
	{
		return ID;
	}
	
	/**
	 * Gets the implicit module for this drive. This method is primarily meant for
	 * internal purposes. It should NOT be used outside of the CLARION Library for purposes 
	 * other than reporting the internal state.
	 * @return The implicit module for this drive.
	 */
	public AbstractImplicitModule getImplicitModule ()
	{
		return DriveMod;
	}
	
	/**
	 * Sets the implicit module for this drive. This method should only be used
	 * during initialization.
	 * @param im The implicit module to set for this drive.
	 * @throws InvalidFormatException If the specified implicit module does not 
	 * contain an output chunk with an ID equal to the ID of this drive or the output chunk 
	 * has the correct ID but is not of type DriveStrength.
	 */
	public void setImplicitModule (AbstractImplicitModule im) throws InvalidFormatException
	{
		DS = null;
		for(AbstractOutputChunk o : im.Output.values())
		{
			if(o.getID().equals(ID) && (o instanceof DriveStrength))
			{
				DS = (DriveStrength)o;
				break;
			}
		}
		if(DS == null)
			throw new InvalidFormatException ("The specified implicit module does " +
					"not contain an output chunk with an ID equal to the ID of this drive or " +
					"the output chunk has the correct ID but is not of type DriveStrength.");
		DriveMod = im;
	}
	
	/**
	 * Gets the drive strength object associated with this drive. This method is mainly used during
	 * initialization and does NOT calculate the drive strength for this drive. If you want to 
	 * calculate the drive strength, call the calculateDriveStrength method.
	 * @return The drive strength object
	 */
	public DriveStrength getDriveStrength ()
	{
		return DS;
	}
    
    /**
     * Gets the deficit level.
     * @return The deficit.
     */
    public double getDeficit ()
    {
    	return Deficit.getActivation();
    }
    
    /**
     * Gets the level of the initial deficit. Once the drive has been initialized, the initial deficit
     * cannot be changed.
     * @return The initial deficit.
     */
    public double getInitialDeficit ()
    {
    	return InitialDeficit.getActivation();
    }
    
    /**
     * Sets the deficit. The initial deficit is left unchanged.
     * @param deficit The deficit value to set.
     */
    public void setDeficit (double deficit)
    {
    	if(Deficit == null)
    	{
    		Deficit = new Value (this, deficit);
    	}
    	else
    	{
	    	Deficit.setActivation(deficit);
    	}
    }
    
    /**
     * Sets the deficit change rate.
     * @param changeRate The rate of change to set.
     */
    public void setDeficitChangeRate (double changeRate)
    {
    	DEFICIT_CHANGE_RATE = changeRate;
    }
    
    /**
     * Updates the deficit using the rate of change.
     */
    protected void updateDeficit ()
    {
    	Deficit.setActivation(Deficit.getActivation() * DEFICIT_CHANGE_RATE);
    }
    
    /**
	 * Checks to see if the specified object is a drive and has the same ID as the specified
	 * drive.
	 * @param drive The object to compare to this drive.
	 * @return True if the two drives are have the same ID, otherwise false.
	 */
	public boolean equals (Object drive)
	{
		if (this == drive)
			return true;
		if (!(drive instanceof Drive))
			return false;
		if (((Drive)drive).getID().equals(ID))
		{
				return true;
		}
		else
			return false;
	}
	
	public int hashCode()
	{
		return hash;
	}
	
	/**
	 * Clones the drive. This method should NOT be used outside of the CLARION library as
	 * it does NOT clone the implicit module and each instance of a drive should have its
	 * own implicit module (at least between different instances of a CLARION agent).
	 * @return A copy of this drive.
	 */
	public Drive clone ()
	{
		Drive d = new Drive (ID,InitialDeficit.getActivation(),DriveMod);
		d.DS = DS;
		d.GAIN = GAIN;
		d.BASELINE = BASELINE;
		d.hash = hash;
		return d;
	}
	
	public String toString()
	{
		if(!(ID instanceof Drive))
			return "Drive ID - " + ID.toString() + ": Deficit = " + Deficit.getActivation() + "\n";
		return "Drive ID - ?: Deficit = " + Deficit.getActivation();
	}
}
