package clarion.system;

import java.util.Collection;

/**
 * This class implements a drive equation within CLARION. It extends the AbstractEquation class.
 * <p>
 * <b>Usage:</b>
 * <p>
 * This class implements one option for calculating drive strengths for a drive using:
 * <br>
 * <b>DriveStrength = (DRIVE_GAIN * SYSTEM_GAIN * UNIVERAL_GAIN) * Deficit * Stimulus + Baseline</b>
 * <p>
 * This class can be used as the implicit module within a drive for calculating a drive strength.
 * <p>
 * If you are using the drive equation, the input MUST contain a dimension with an ID 
 * equal to the ID of this drive associated with this equation that contains values with the IDs: 
 * <ul>
 * <li><b>DRIVE_GAIN</b></li>
 * <li><b>SYSTEM_GAIN</b></li>
 * <li><b>UNIVERAL_GAIN</b></li>
 * <li><b>STIMULUS</b></li>
 * <li><b>DEFICIT</b></li>
 * <li><b>BASELINE</b></li>
 * </ul>
 * The above IDs must be of enumerated type <b>RequiredEquationInputs</b>. All other dimensions and values in the 
 * input will be ignored.
 * <p>
 * Note that if the drive, with which this equation is associated, belongs to both the BIS and BAS, the SYSTEM_GAIN should be set
 * to the average of the gains of the two drives.
 * <p>
 * While it is completely within the capabilities of this system to use equations within CLARION, equations are NOT sub-symbolic or 
 * distributed in nature. Therefore, it is encouraged that you only use equations in the bottom level for testing and debugging 
 * purposes. Instead, you are advised to use a more sub-symbolic structure (such as a neural network) for the implicit modules
 * within CLARION.
 * @version 6.0.4.5
 * @author Nick Wilson
 */
public class DriveEquation extends AbstractEquation {
	/**The enumerator to use for the IDs for the inputs that are required for this equation.*/
	public enum RequiredEquationInputs {DRIVE_GAIN, SYSTEM_GAIN, UNIVERSAL_GAIN, STIMULUS, 
		DEFICIT, BASELINE};
		
	/**The ID of the drive to which this equation is attached.*/
	private Object ID;
	
	/**
	 * Initializes a drive equation.
	 * @param InputSpace The input space for the equation.
	 * @param Outputs The outputs for the equation.
	 */
	public DriveEquation(Object id, Collection<Dimension> InputSpace, AbstractOutputChunkCollection<? extends AbstractOutputChunk> Outputs) {
		super(InputSpace, Outputs);
		ID = id;
	}

	/**
     * Calculates the output activations based on the current input. This method should not 
     * be called until the setInput method has been called setting the input activations to 
     * the current state.
     * @throws InvalidFormatException If the input to the equation does not contain the required
     * equation inputs.
     */
	public void forwardPass() throws InvalidFormatException{
		Value s = InputAsCollection.get(ID).get(RequiredEquationInputs.STIMULUS);
		Value sgain = InputAsCollection.get(ID).get(RequiredEquationInputs.SYSTEM_GAIN);
		if(s == null || sgain == null)
			throw new InvalidFormatException ("This drive is using the equation and "+ 
					"while the drive input contains a dimension with the same ID as this drive, "+
					"that dimension does not contain either a value with the ID: STIMULUS "+
					"(of enumerated type RequiredEquationInputs) or a value with the ID: SYSTEM_GAIN " +
					"(of enumerated type RequiredEquationInputs).");
		Value dgain = InputAsCollection.get(ID).get(RequiredEquationInputs.DRIVE_GAIN);
		Value ugain = InputAsCollection.get(ID).get(RequiredEquationInputs.UNIVERSAL_GAIN);
		Value d = InputAsCollection.get(ID).get(RequiredEquationInputs.DEFICIT);
		Value b = InputAsCollection.get(ID).get(RequiredEquationInputs.BASELINE);
		AbstractOutputChunk ds = Output.get(ID);
		ds.setActivation((dgain.getActivation() * sgain.getActivation() * ugain.getActivation())
				* d.getActivation() * s.getActivation() + b.getActivation());
	}
	
	/**
	 * This method does nothing as it is not used by the CLARION Library for this equation.
	 * @return False
	 */
    public boolean checkMatchCriterion()
    {
    		return false;
    }

}
