package clarion.system;

import java.util.Collection;
import java.util.Map;

/**
 * This class implements a generalized condition chunk within CLARION. It extends the DimensionValueCollection class.
 * <p>
 * <b>Usage:</b>
 * <p>
 * A condition usually contains the same dimension-value pairs as (or a same set of) the input nodes of an implicit module
 * and generally has the same structure as the bottom level to which it is related (or from which it was extracted).
 * <p>
 * Values within a dimension in a condition are OR linked; whereas dimensions themselves are AND linked. This means that
 * multiple values within the same dimension can have activations greater than 0. This is different than the sensory input
 * (or a dimension-value collection in general), which can only have at most 1 activated value for any dimension.
 * <p>
 * Note that while we refer to the condition of a rule as a "chunk", it is not the same as a chunk as defined
 * by the CLARION theory (as well as the AbstractChunk class within this library). The condition chunk is
 * a specialized kind of chunk in that it does not have an ID, and it has special methods for calculating its
 * strength when compared to another dimension-value collection (usually the current input).
 * @version 6.0.4
 * @author Nick Wilson
 */
public class GeneralizedConditionChunk extends DimensionValueCollection{
	private static final long serialVersionUID = 5496821003835063531L;
	
	/**
	 * Initializes a condition.
	 */
	public GeneralizedConditionChunk ()
	{
		super();
	}
	
	/**
	 * Initializes a condition with the collection of dimensions specified.
	 * @param dims The dimensions for the collection.
	 */
	public GeneralizedConditionChunk (Collection <? extends Dimension> dims)
	{
		super(dims);
	}
	
	/**
	 * Initializes the condition with the map of dimensions.
	 * @param map The map of dimensions for the collection.
	 */
	public GeneralizedConditionChunk (Map <? extends Object, ? extends Dimension> map)
	{
		super(map);
	}
	
	/**
	 * Gets the weight of the condition. By default this is 1/n where n is the number of dimensions.
	 * @return The weight of the weight
	 */
	public double getWeight()
	{
		return 1/(double)size();
	}
	
	/**
	 * Gets the strength of the condition based on the current input.
	 * @param CurrentInput The current input.
	 * @return The strength of the condition.
	 */
	public double getStrength(Collection <? extends Dimension> CurrentInput)
	{
		if(CurrentInput == null)
			return 0;
		double Strength = 0;
		double Weight = getWeight();
		for(Dimension i : CurrentInput)
		{
			Dimension k = get(i.getID());
			if(k != null)
			{
				if(k.checkMatchAll())
					Strength += k.getMaxActivation() * Weight;
				else
				{
					double Max = 0;
					for (Value j : i.values())
					{
						Value l = k.get(j.getID());
						if(l != null && l.getActivation() >= j.getActivation() && 
								j.getActivation() >  Max)
						{
							Max = j.getActivation();
						}
					}
					Strength += Max * Weight;
				}
			}
		}
		return Strength;
	}
	
	/**
	 * Checks to see if the specified condition is covered by this condition.
	 * @param cond The condition to check against this condition.
	 * @return True if this condition covers the specified condition, otherwise false.
	 */
	public boolean covers (GeneralizedConditionChunk cond)
	{
		if(cond == null)
			return false;
		
		for(Dimension i : cond.values())
		{
			if(containsKey(i.getID()))
			{
				Dimension d = get(i.getID());
				if(!d.checkMatchAll())
				{
					for(Value j : i.values())
					{
						if(!(d.containsKey(j.getID())))
							return false;
						if(!d.get(j.getID()).equals(j) && !d.get(j.getID()).isFullyActivated())
							return false;
					}
				}
			}
			else
				return false;
		}
		return true;
	}
	
	public boolean checkMatchAll()
	{
		boolean ma = true;
		for(Dimension dd : this.values())
		{
			if(!dd.checkMatchAll())
				ma = false;
		}
		return ma;
	}
	
	/**
	 * This method uses a special hash function based on the activations of the values within the condition.
	 * The hash is created by summing the hashes of all the values within the condition multiplied by the
	 * activation of those values. Written mathematically, it looks like:
	 * <br>
	 * <b>Hash = Sum(ValueHash * ValueActivation)</b>
	 * <p>
	 * If the condition is empty (i.e. it has no dimension-value pairs, which is an option in fixed rules), this 
	 * method will return the default system hash code.
	 * @return The hash code for the condition.
	 */
	public int hashCode()
	{
		if(size() == 0)
			return System.identityHashCode(this);
		float dhash = 0;
		for(Value v : getValueCollection())
		{
			dhash += v.hashCode() * v.getActivation();
		}
		return Math.round(dhash);
	}
	
	/**
	 * Clones the condition (including all of it's dimensions).
	 * @return A copy of the condition.
	 */
	public GeneralizedConditionChunk clone ()
	{
		GeneralizedConditionChunk a = new GeneralizedConditionChunk();
		for (Dimension i : values())
		{
			Dimension d = i.clone();
			a.put(d.getID(), d);
		}
		return a;
	}
}
