package clarion.system;

import java.util.Collection;

/**
 * This class implements a goal selection equation within CLARION. It extends the AbstractEquation class.
 * <p>
 * <b>Usage:</b>
 * <p>
 * This class implements one option for setting a goal using the goal equation:
 * <br>
 * <b>GoalActivation = Sum(RelevantDriveStrength * Relevance)</b>
 * <p>
 * This class can be used within the goal selection module of the MCS for setting the goal in the goal structure.
 * <p>
 * The inputs to the equation are a collection of "DriveStrengths" combined into a dimension-value collection obtained from the 
 * toDimensionValueCollection method in the DriveStrengthCollection class and the relevance measures obtained from the goals. 
 * The outputs are Goals.
 * <p>
 * While it is completely within the capabilities of this system to use equations within CLARION, equations are NOT sub-symbolic or 
 * distributed in nature. Therefore, it is encouraged that you only use equations in the bottom level for testing and debugging 
 * purposes. Instead, you are advised to use a more sub-symbolic structure (such as a neural network) for the implicit modules
 * within CLARION.
 * @version 6.0.4
 * @author Nick Wilson
 */
public class GoalSelectionEquation extends AbstractEquation {
	
	/**
	 * Initializes the goal selection equation.
	 * @param InputSpace The input space for the goal selection equation, which is a dimension-value
	 * collection of drive strengths.
	 * @param Outputs The outputs for the equation (i.e. goals).
	 */
	public GoalSelectionEquation(Collection<Dimension> InputSpace, 
			AbstractOutputChunkCollection<? extends AbstractOutputChunk> Outputs) {
		super(InputSpace, Outputs);
	}

	public void forwardPass() {
		for(AbstractOutputChunk g : Output.values())
		{
			g.resetActivation();
			for(Value ds : InputAsCollection.getValueCollection())
			{
				if(((Goal)g).getRelevance(ds.getID()) != null)
				{
					g.setActivation(g.getActivation() + 
							((Goal)g).getRelevance(ds.getID()).getActivation()
							* ds.getActivation());
				}
			}
		}
	}
	
	/**
	 * This method does nothing as it is not used by the CLARION Library for this equation.
	 * @return False
	 */
	public boolean checkMatchCriterion() {
		return false;
	}

}
