package clarion.system;

import java.util.Collection;

/**
 * This class implements a goal selection module within CLARION.
 * <p>
 * <b>Usage:</b>
 * <p>
 * This module can be optionally added to the MCS if the user wishes to have the MCS handle goal setting. 
 * This module uses an implicit module to determine the goal activation level. 
 * The user can decide, during initialization, the implicit module they wish to use.
 * <p>
 * If this module is initialized without specifying an implicit module to use, the GoalSelectionEquation 
 * will be used.
 * @version 6.0.5
 * @author Nick Wilson
 */
public class GoalSelectionModule {
	
	/**The implicit representation used for goal selection.*/
	private AbstractImplicitModule GoalSelectionImplicitModule;
	
	/**The goal structure.*/
	private GoalStructure GS;
	/**The stochastic selector for goal selection.*/
	public StochasticSelector Selector = new StochasticSelector();
	
	/**
	 * Initializes the goal selection module to use the goal selection equation. The goal selection equation 
	 * will select from the specified goals using the specified drive strengths.
	 * @param RelevantDrives The drive strengths for relevant drives used to make goal selections.
	 * @param Goals The goals from which to select.
	 */
	public GoalSelectionModule (Collection <DriveStrength> RelevantDrives, Collection <Goal> Goals)
	{
		GoalCollection gs = new GoalCollection (Goals);
		DriveStrengthCollection dsc = new DriveStrengthCollection (RelevantDrives);
		GoalSelectionImplicitModule = new GoalSelectionEquation (dsc.toDimensionValueCollection().values(), gs);
	}
	
	/**
	 * Initializes the goal selection module given the specified implicit module.
	 * <p>
     * Note that the specified implicit module MUST have output chunks each of 
     * which is a possible goals in the goal structure.
     * <p>
     * The goal selection implicit module will only select over the possible goals 
     * that are specified as an output chunk on the output of the specified implicit 
     * module.
     * <p>
     * If the goal selection implicit module contains any output nodes that are of 
     * a type other than Goal, this constructor will throw an exception.
	 * @param im The goal selection implicit module to set for this module.
	 * @throws InvalidFormatException If the output layer contains nodes of a type other than
	 * Goal.
	 */
	public GoalSelectionModule (AbstractImplicitModule im) throws InvalidFormatException
	{
		for(AbstractOutputChunk o : im.Output.values())
			if(!(o instanceof Goal))
				throw new InvalidFormatException ("The output layer of the specified implicit module " +
						"contains nodes that are of a type other than Goal.");
		GoalSelectionImplicitModule = im;
	}
	
	/**
	 * Selects a goal from all of the possible goals tied to the output layer of implicit module in this 
	 * module using the specified collection of drive strengths.
	 * <p>
	 * If a new goal is selected that is different than the currently chosen goal then the currently chosen 
	 * goal is removed from the goal structure and the new goal is set as the currently chosen goal. If the goal 
	 * structure has not been initialized or does not contain any possible goals, this method does nothing.
	 * @param DriveStrengths The drive strengths of relevant drives.
	 * @param TimeStamp The current time stamp.
	 */
	public void setGoal (DriveStrengthCollection DriveStrengths, long TimeStamp) 
	{
		if(GS == null || GS.getPossibleGoals() == null)
			return;
		
		GoalSelectionImplicitModule.setInput(DriveStrengths.toDimensionValueCollection());
		GoalSelectionImplicitModule.forwardPass();
		
		if(GS.getCurrentGoal() != null)
		{
			for(Goal g : GS.getPossibleGoals())
			{
				g.setActivation(GoalSelectionImplicitModule.getOutput(g.getID()).getActivation());
				if(GS.getCurrentGoal().equalsID(g))
					g.setBLSelectionMeasure(g.adjustSelectionMeasure(g.getActivation(), g.FULL_ACTIVATION_LEVEL));
				else
					g.setBLSelectionMeasure(g.adjustSelectionMeasure(g.getActivation(), g.MINIMUM_ACTIVATION_THRESHOLD));
			}
		}
		GS.remove(GS.getCurrentGoal());
		Goal newgoal = (Goal)Selector.select(GS.getPossibleGoals());
		GS.add(newgoal);
		newgoal.addTimeStamp(TimeStamp);
	}
	
	/**
	 * Gets the implicit module used by this module. This method is primarily meant for
	 * internal purposes. It should NOT be used outside of the CLARION Library for purposes 
	 * other than reporting the internal state.
	 * @return The implicit module of this module.
	 */
	public AbstractImplicitModule getImplicitModule ()
	{
		return GoalSelectionImplicitModule;
	}
	
	/**
	 * Attaches the specified goal structure to this module.
	 * @param gs The goal structure.
	 */
	public void attachGoalStructure (GoalStructure gs)
	{
		GS = gs;
	}
}
