package clarion.system;

import java.util.Collection;
import java.util.ListIterator;

/**
 * This class implements the goal structure within CLARION. It extends the AbstractIntermediateModule class.
 * Conceptually, this system exists on the top level of the MS. However,
 * implementationally it is logical to specify it as an intermediate module since it is used
 * by several subsystems. Therefore, in the CLARION Library, the goal structure is contained within the
 * CLARION class.
 * <p>
 * <b>Usage:</b>
 * <p>
 * The goal structure's main function is to act as a container for the goals within CLARION. In addition to 
 * containing methods for adding and removing items from the goal list, it also keeps track of all of the
 * possible goals that can be set and has a method for getting the current goal.
 * <p>
 * It is EXTREMELY important that you only manipulate the goal structure using the add and remove
 * methods that have been overridden by this class.
 * <p>
 * Currently this is simply a container for the goal list.
 * <p>
 * This class contains both global (static) and local constants. The default is to use the local 
 * constants. If you want to change any of the global constants, you need to do so before any
 * instances of this class are initialized.
 * @version 6.0.4
 * @author Nick Wilson
 */
public class GoalStructure extends AbstractIntermediateModule <Goal>{
	private static final long serialVersionUID = -3073724626285378374L;
	
	/**The number of slots in the goal list.*/
	public static int GLOBAL_GOAL_LIST_CAPACITY = 7;
	/**The number of slots in the goal list.*/
	public int GOAL_LIST_CAPACITY = GLOBAL_GOAL_LIST_CAPACITY;
	/**All of the possible Goals that can be added to the goal structure.*/
	private GoalCollection PossibleGoals;
	
	/**
	 * Initializes the goal structure. During initialization this instance of the goal
	 * structure will attach itself to the CLARION agent you specify.
	 * @param Agent The agent to which the goal structure is being attached.
	 */
	public GoalStructure (CLARION Agent) 
	{
		super(Agent);
	}
	
	/**
	 * Gets the current goal. The current goal is the goal in the list with the highest BLA. If no
	 * goals are in the goal list, this method returns null.
	 * @return The current goal. Null if no goals are in the list.
	 */
	public Goal getCurrentGoal ()
	{
		if(size() != 0)
		{
			ListIterator <Goal> i = listIterator();
			Goal current = i.next();
			while(i.hasNext())
			{
				Goal check = i.next();
				if(check.getBLA() > current.getBLA())
					current = check;
			}
			return current;
		}
		return null;
	}
	
	/**
	 * Adds a goal to the goal list. If you are trying to add a goal to the goal structure that has
	 * dimension-value pairs that do not exist in CLARION's internal input space or the goal structure
	 * is full, this method will throw an exception.
	 * <p>
	 * If you want to add a new goal to the list of possible goals in the goal structure, call the 
	 * addPossibleGoal method.
	 * @param G The goal to add.
	 * @return True only if the goal was successfully added.
	 * @throws FullContainerException If the goal structure is full.
	 * @throws IllegalArgumentException If the input space does not contain all of the dimension-value
	 * pairs specified in the goal.
	 */
	public boolean add (Goal G) throws FullContainerException, IllegalArgumentException
	{
		if(size() < GOAL_LIST_CAPACITY)
		{
			for(Dimension d : G.values())
			{
				Dimension a = InputSpace.get(d.getID());
				if(a == null)
					throw new IllegalArgumentException ("The specified goal contains a dimension " +
							"that has not been specified within the agent's internal input space.");
				for(Value v : d.values())
				{
					if(!a.containsKey(v.getID()))
						throw new IllegalArgumentException ("The specified goal contains a value " +
								"in dimension " + d.getID().toString() + " that has not been specified " +
								"within the agent's internal input space.");
				}
			}
			
			G.activateGoal();
			super.add(G);
		}
		else
			throw new FullContainerException();
		return true;
	}
	
	/**
	 * Adds a collection of goals to the goal structure.
	 * @param G The goals to add.
	 * @return True only if the goals were successfully added.
	 */
	public boolean addAll (Collection <? extends Goal> G)
	{
		for(Goal g : G)
			add(g);
		return true;
	}
	
	/**
	 * Removes a goal from the goal list.
	 * @param Index The index of the goal to remove.
	 * @return The removed goal.
	 */
	public Goal remove (int Index)
	{
		Goal g = super.remove(Index);
		g.deactivateGoal();
		return g;
	}
	
	/**
	 * Removes the goal object from the goal list.
	 * @param G The goal to remove.
	 * @return True if the goal was successfully removed.
	 */
	public boolean remove (Object G)
	{
		super.remove(G);
		if(G instanceof Goal)
			((Goal)G).deactivateGoal();
		return true;
	}
	
	/**
	 * Gets the number of possible goals.
	 * @return The number of possible goals.
	 */
	public int getNumPossibleGoals ()
	{
		if(PossibleGoals != null)
			return  PossibleGoals.size();
		else
			return 0;
	}
	
	/**
	 * Gets a collection of the possible goals.
	 * @return An unmodifiable collection of goals.
	 */
	public Collection <Goal> getPossibleGoals ()
	{
		if(PossibleGoals != null)
			return PossibleGoals.values();
		return null;
	}
	
	/**
	 * Adds a goal to the list of possible goals. If the goal is already in the list of 
	 * possible goals, it will not be added.
	 * <p>
	 * Remember that the input layer of an implicit module or rule in the ACS must contain nodes for all dimension-value pairs 
	 * contained within any of the possible goals you wish to have as input into the network or rule condition.
	 * @param G The possible goal to add.
	 * @return True if the goal was successfully added.
	 */
	public boolean addPossibleGoal (Goal G)
	{
		if(PossibleGoals == null)
		{
			PossibleGoals = new GoalCollection ();
		}
		
		if(!PossibleGoals.containsKey(G.getID()))
		{
			Goal g = G.clone();
			updateInputSpace(g.values());
			PossibleGoals.put(g.getID(),g);
			return true;
		}
		return false;
	}
	
	/**
	 * Adds a collection of goals to the list of possible goals. If a goal from the collection 
	 * is already in the list of possible goals, it will not be added.
	 * <p>
	 * Remember that the input layer of an implicit module or rule in the ACS must contain nodes for all dimension-value pairs 
	 * contained within any of the possible goals you wish to have as input into the network or rule condition.
	 * @param Goals The collection of possible goals to add.
	 * @return True if any of the goals were successfully added.
	 */
	public boolean addPossibleGoals (Collection <Goal> Goals)
	{
		boolean rettrue = false;
		for(Goal g : Goals)
		{
			if(addPossibleGoal(g))
				rettrue = true;
		}
		return rettrue;
	}
	
	/**
	 * Attaches the goal structure to the specified CLARION agent.
	 * @param Agent The agent to wish this goal structure will be attached.
	 */
	protected void attachSelfToAgent (CLARION Agent)
	{
		Agent.attachGoalStructure(this);
	}
}
