package clarion.system;

import java.util.Collection;
import java.util.Iterator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

/**
 * This class implements a hidden rule collection within CLARION. It extends the HashMap class.
 * This class is used exclusively within a rule collection for tracking child rules and rule
 * variations.
 * <p>
 * <b>Usage:</b>
 * <p>
 * The hidden rule collection acts as a container for rules that are not explicitly accessible by a
 * CLARION agent. That is, the rules contained within a hidden rule collection can't be used for
 * action decision making.
 * <p>
 * Rules that are normally stored in the hidden rule collection include:<br>
 * <ul>
 * <li><b>Rule variations</b> - rules that differ by a single value within the condition to rules currently in 
 * the rule store. They act as minor variations whose match statistics are tracked and, based on a information 
 * gain measure (see the Sun tutorial, 2003 or the RuleRefiner class in this library), these variations have a 
 * chance of becoming actual rules within the rule store acting as a more generalized and/or specialized version 
 * of the rule in the rule store to which it is associated.</li>
 * <li><b>Child rules</b> - rules that USED to be in the rule store, but have been replaced by more generalized rules
 * that cover them. These rules are maintained in the event that the more generalized rule is deleted from the
 * rule store (for various reasons), at which time they have the opportunity to return back to the rule store.</li>
 * </ul>
 * <p>
 * The hidden rule collection itself is a somewhat more simplified version of a rule collection in that it
 * does not need to track (or generate) things like rule variations, child rules, or collection-wide match statistics.
 * In addition, it doesn't have to worry about correctly placing newly added rules within the collection by determining
 * if the rule should be in the collection or within the child collection.
 * <p>
 * Conceptually speaking, rules contained within a hidden rule collection really only exist "abstractly."
 * It is important to reiterate that rules contained within this collection are NOT accessible to the CLARION agent
 * for any purposes other than rule refinement.
 * @version 6.0.4.5
 * @author Nick Wilson
 */
public class HiddenRuleCollection extends HashMap<Object, HashMap<Integer,AbstractRule>>{
	private static final long serialVersionUID = 8247993623873488699L;
	
	/**
	 * Initializes a rule collection.
	 */
	public HiddenRuleCollection ()
	{
		super();
	}
	
	/**
	 * Initializes a rule collection with the collection of rules specified.
	 * @param rules The rules for the collection.
	 */
	public HiddenRuleCollection (Collection <? extends AbstractRule> rules)
	{
		super();
		for(AbstractRule r : rules)
			put(r);
	}
	
	/**
	 * Gets a rule in the collection that matches the specified condition and 
	 * action. If the collection does not contain a rule with the specified condition and 
	 * action, this method returns null.
	 * @param cond The condition of the rule to get.
	 * @param act The action of the rules to get.
	 * @return A rules with the specified condition and action.
	 */
	public AbstractRule get (GeneralizedConditionChunk cond, AbstractAction act)
	{
		HashMap<Integer,AbstractRule> rc = get(act.getID());
		if (rc == null)
			return null;
		return rc.get(cond.hashCode());
	}
	
	/**
	 * Checks to see if the collection contains the specified rule.
	 * @param rule The rule object you wish to check for within this collection.
	 * @return True if the collection contains the rule specified, otherwise false.
	 */
	public boolean contains (AbstractRule rule)
	{
		if(get(rule.rCondition, rule.rAction) != null)
			return true;
		return false;
	}

	/**
	 * This method does nothing. It has been overridden from HashMap to return null
	 * without performing an operation. Instead, use the overloaded put method that takes
	 * in an abstract rule as a parameter.
	 * @param key
	 * @param map
	 * @return null
	 */
	public HashMap <Integer, AbstractRule> put (Object key, HashMap <Integer,AbstractRule> map)
	{
		return null;
	}
	
	/**
	 * This method does nothing. It has been overridden from HashMap return without 
	 * performing an operation. Instead, use the overloaded putAll method that takes
	 * in a collection of abstract rules.
	 * @param map
	 */
	public void putAll (Map<? extends Object, ? extends HashMap <Integer,AbstractRule>> map)
	{
		return;
	}
	
	/**
	 * This method does nothing. It has been overridden from HashMap to return null
	 * without performing an operation. Instead, use the overloaded remove method that takes
	 * in a condition and action.
	 * @param key
	 * @return null
	 */
	public HashMap<Integer, AbstractRule> remove (Object key)
	{
		return null;
	}
	
	/**
	 * Puts a rule into the rule collection. If the specified rule is equal to any rules 
	 * currently in the collection, the rule will NOT be added. In addition, this method
	 * also removes all of the rule variations from the specified rule (if the rule is
	 * refineable) since it is not necessary to track variations for rules that are not in
	 * the rule store.
	 * @param R The rule to add.
	 */
	public void put (AbstractRule R)
	{
		HashMap <Integer, AbstractRule> rc = get(R.getAction().getID());
		if(rc == null)
		{
			rc = new HashMap<Integer, AbstractRule> ();
			rc.put(R.getCondition().hashCode(), R);
			super.put(R.getAction().getID(), rc);
		}
		else
		{
			if(!rc.containsKey(R.getCondition().hashCode()))
			{
				rc.put(R.getCondition().hashCode(), R);
			}
		}
		if(R instanceof RefineableRule && ((RefineableRule)R).getVariations() != null)
			((RefineableRule)R).resetVariations();
	}
	
	/**
	 * Puts a collection of rules into the rule collection. If a rule in the specified 
	 * collection is equal to any rules currently in this collection, the rule will NOT 
	 * be added.
	 * @param R The rules to put in the collection.
	 */
	public void putAll (Collection <? extends AbstractRule> R)
	{
		for(AbstractRule r : R)
		{
			put(r);
		}
	}
	
	/**
	 * Removes the rule from this rule collection with the specified condition and action.
	 * If the rule collection does not contain a rule with the specified condition and action,
	 * this method returns null.
	 * @param cond The condition of the rule you want to remove.
	 * @param act The action of the rule you want to remove.
	 * @return The rule that was removed. If no rule was found in the collection, then null.
	 */
	public AbstractRule remove (GeneralizedConditionChunk cond, AbstractAction act)
	{
		HashMap<Integer,AbstractRule> rc = get(act.getID());
		if (rc == null)
			return null;
		AbstractRule r = rc.remove(cond.hashCode());
		if(rc.size() == 0)
			super.remove(act.getID());
		return r;
	}
	
	/**
	 * Gets all of the rules in the rule collection.
	 * @return A collection containing the rules in the rule collection.
	 */
	public Collection <AbstractRule> getRules()
	{
		LinkedList <AbstractRule> rc = new LinkedList<AbstractRule> ();
		for(HashMap<Integer, AbstractRule> i : values())
		{
			rc.addAll(i.values());
		}
		return rc;
	}
	
	/**
	 * Discounts the positive and negative match statistics for all rules in the rule 
	 * collection.
	 * @param Discount The discount factor to be applied to the match statistics of the rules.
	 */
	public void discountMatchStatistics(double Discount)
	{
		for(HashMap <Integer,AbstractRule> r : values())
		{
			for(AbstractRule i : r.values())
			{
				i.setPM(i.getPM() * Discount);
				i.setNM(i.getNM() * Discount);
			}
		}
	}
	
	/**
	 * Updates the positive or negative match statistics based on the information specified.
	 * @param CurrentInput The current input.
	 * @param ChosenAction The chosen action.
	 * @param feedback The feedback.
	 * @param TimeStamp	The current time stamp.
	 */
	public void updateMatchStatistics(DimensionValueCollection CurrentInput, AbstractAction ChosenAction, 
			double feedback, AbstractMatchCalculator MatchCalculator, long TimeStamp)
	{
		HashMap<Integer,AbstractRule> rc = get(ChosenAction.getID());
		if(rc != null)
		{
			for(AbstractRule i : rc.values())
			{
				i.setCurrentInput(CurrentInput);
				if(i.checkMatchCriterion(TimeStamp))
				{
					i.setFeedback(feedback);
					if(i instanceof InterfaceHasMatchCalculator)
						i.updateMatchStatistics(((InterfaceHasMatchCalculator)i).getMatchCalculator());
					else
						i.updateMatchStatistics(MatchCalculator);
				}
			}
		}
	}
	
	/**
	 * Returns the number of rules in the rule collection.
	 */
	public int size ()
	{
		return getRules().size();
	}
	
	public String toString()
	{
		String s = "";
		for(Iterator <AbstractRule> i = getRules().iterator(); i.hasNext();)
		{
			AbstractRule r = i.next();
			s += r.toString();
			if(i.hasNext())
				s+="\n";
		}
		return s;
	}
}
