package clarion.system;

/**
 * This class implements the Meta Cognitive Subsystem (MCS) within CLARION. 
 * It extends the AbstractSubsystem class.
 * <p>
 * <b>Usage:</b>
 * <p>
 * This class contains the modules of the MCS along with methods for getting and setting those modules.
 * The user can decide, as desired, which modules to attach to the MCS and the CLARION Library will
 * use those modules if they are attached, otherwise it wont. In this sense, the CLARION Library
 * is dynamic and flexible about which modules appear within the MCS.
 * <p>
 * The modules of the MCS are accessed most commonly by the CLARION class through the perceive and act methods.
 * Note that if a developer wished to write a currently unimplemented module into the MCS, the only ways
 * to do so would be to either edit the CLARION library (not recommended) or write their own
 * MCS and CLARION classes by extending those classes and then adding/overriding (respectively) the new 
 * modules/perceive and act functions (again respectively).
 * <p>
 * Currently, the following modules have been implemented, and can be attached, as needed, within the MCS:
 * <ul>
 * <li>The ACS Level Probability Setting Module</li>
 * <li>The Goal Selection Module </li>
 * </ul>
 * <p>
 * In addition, it is not required that a CLARION agent even have an MCS at all. If the task being performed
 * by the agent does not require an MCS, the user can choose simply not to attach the MCS to the
 * CLARION agent, and the agent will be able to operate quite successfully without this subsystem being
 * specified (for some tasks).
 * @version 6.0.4
 * @author Nick Wilson
 */
public class MCS extends AbstractSubsystem{
	
	/**The goal selection module.*/
	private GoalSelectionModule GoalSelectionModule;
	
	/**The ACS level probability setting module*/
	private ACSLevelProbabilitySettingModule ACSLevelProbabilitySettingModule;
	
	/**
	 * Minimally initializes the MCS. This constructor builds a "frame" for this instance of
	 * the MS from which all desired components can be attached. During initialization this
	 * instance of the MCS will attach itself to the CLARION agent you specify.
	 * @param Agent The agent to which the MCS is being attached.
	 */
	public MCS (CLARION Agent) 
	{
		super(Agent);
	}
	
	/**
	 * Attaches the goal selection module.
	 * @param module The module to attach.
	 */
	public void attachGoalSelectionModule (GoalSelectionModule module)
	{
		GoalSelectionModule = module;
	}
	
	/**
	 * Attaches the ACS level probability setting module.
	 * @param module The module to attach.
	 */
	public void attachACSLevelProbabilitySettingModule (ACSLevelProbabilitySettingModule module)
	{
		ACSLevelProbabilitySettingModule = module;
	}
	
	/**
	 * Gets the goal selection module.
	 * @return The goal selection module.
	 */
	public GoalSelectionModule getGoalSelectionModule ()
	{
		return GoalSelectionModule;
	}
	
	/**
	 * Gets the ACS level probability setting module.
	 * @return The ACS level probability setting module.
	 */
	public ACSLevelProbabilitySettingModule getACSLevelProbabilitySettingModule ()
	{
		return ACSLevelProbabilitySettingModule;
	}
	
	/**
	 * Attaches the MCS to the specified CLARION agent.
	 * @param Agent The agent to wish this MCS will be attached.
	 */
	protected void attachSelfToAgent(CLARION Agent) {
		Agent.attachMCS(this);
	}

	/**
	 * Performs the appropriate end of episode instructions for the MCS. This method is called by the
	 * CLARION class when its endEpisode method is called.
	 */
	protected void endEpisode() {
		return;
	}
}