package clarion.system;

import java.util.Collection;
import java.util.Iterator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

/**
 * This class implements a rule collection within CLARION. It extends the HashMap class and
 * implements  the InterfaceTracksMatchStatistics and InterfaceHandlesFeedback interfaces.
 * <p>
 * <b>Usage:</b>
 * <p>
 * The rule collection acts as a container for rules that are used by a CLARION agent for
 * action decision making. This class is most often used within the ACS as the container for the 
 * rule stores.
 * <p>
 * The rule collection also tracks (or generates): rule variations, child rules, and collection-wide match statistics.
 * In addition, it has the ability to correctly place newly added rules within the collection by determining
 * if the rule should be in the collection or within a child collection.
 * <p>
 * For all of the rules in the rule collection that are refineable (i.e. they extend from the RefineableRule class), 
 * hidden minor variations of those rules are generated automatically and tracked by this class in order to 
 * facilitate generalization and specialization.
 * <p>
 * In addition to tracking match statistics for the hidden variations, the rule collection also keeps track of
 * collection-wide match statistics that are used to for variable level selection within the ACS as well as for
 * performance monitoring and reporting purposes.
 * <p>
 * This class contains both global (static) and local constants. The default is to use the local 
 * constants. If you want to change any of the global constants, you need to do so before any
 * instances of this class are initialized.
 * @version 6.0.4.5
 * @author Nick Wilson
 */
public class RuleCollection extends HashMap<Object, HashMap<Integer,AbstractRule>> implements InterfaceTracksMatchStatistics, InterfaceHandlesFeedback{
	private static final long serialVersionUID = 8247993623873488699L;
	
	/**Contains the hidden rule variations for all rules in the rule collection.*/
	protected HiddenRuleCollection Variations = new HiddenRuleCollection ();
	/**Contains any existent child rules for the rules in the rule collection.*/
	protected HiddenRuleCollection Children = new HiddenRuleCollection ();
	
	/**The immediate feedback (if given).*/
	private double Feedback;
	/**The positive match counter.*/
	protected double PM = 0;
	/**The negative match counter.*/
	protected double NM = 0;
	/**The threshold that must be passed to meet the positive match criterion.*/
	public static double GLOBAL_POSITIVE_MATCH_THRESHOLD = .9;
    /**The threshold that must be passed to meet the positive match criterion.*/
    public double POSITIVE_MATCH_THRESHOLD = GLOBAL_POSITIVE_MATCH_THRESHOLD;
	
	/**
	 * Initializes a rule collection.
	 */
	public RuleCollection ()
	{
		super();
	}
	
	/**
	 * Initializes a rule collection with the collection of rules specified.
	 * @param rules The rules for the collection.
	 */
	public RuleCollection (Collection <? extends AbstractRule> rules)
	{
		super();
		for(AbstractRule r : rules)
			put(r);
	}
	
	/**
	 * Gets a rule in the collection that matches the specified condition and 
	 * action. If the collection does not contain a rule with the specified condition and 
	 * action, this method returns null.
	 * @param cond The condition of the rule to get.
	 * @param act The action of the rules to get.
	 * @return A rules with the specified condition and action.
	 */
	public AbstractRule get (GeneralizedConditionChunk cond, AbstractAction act)
	{
		HashMap<Integer,AbstractRule> rc = get(act.getID());
		if (rc == null)
			return null;
		return rc.get(cond.hashCode());
	}
	
	/**
	 * Gets a rule in the collection that covers the specified condition and 
	 * action. If the collection does not contain a rule that covers the specified condition and 
	 * action, this method returns null.
	 * <p>
	 * This method does NOT guarantee that the rule returned is the only rule in the collection that
	 * covers the specified condition and action.
	 * @param cond The condition the rule to get must cover.
	 * @param act The action of the rules to get.
	 * @return A rules that covers the specified condition and action.
	 */
	public AbstractRule getCover (GeneralizedConditionChunk cond, AbstractAction act)
	{
		HashMap<Integer,AbstractRule> rc = get(act.getID());
		if(rc != null)
		{
			for(AbstractRule r : rc.values())
			{
				if(r.getCondition().covers(cond))
				{
					return r;
				}
			}
		}
		return null;
	}
	
	/**
	 * Checks to see if the collection contains the specified rule.
	 * @param rule The rule object you wish to check for within this collection.
	 * @return True if the collection contains the rule specified, otherwise false.
	 */
	public boolean contains (AbstractRule rule)
	{
		if(get(rule.rCondition, rule.rAction) != null)
			return true;
		return false;
	}
	
	/**
	 * Checks to see if the collection contains a rule that covers the specified rule.
	 * @param rule The rule object you wish to check within this collection.
	 * @return True if the collection contains a rule that covers the rule specified, 
	 * otherwise false.
	 */
	public boolean containsCover (AbstractRule rule)
	{
		if(getCover(rule.rCondition, rule.rAction) != null)
			return true;
		return false;
	}
	
	/**
	 * This method does nothing. It has been overridden from HashMap to return null
	 * without performing an operation. Instead, use the overloaded put method that takes
	 * an AbstractRule.
	 * @param key
	 * @param map
	 * @return null
	 */
	public HashMap <Integer, AbstractRule> put (Object key, HashMap <Integer, AbstractRule> map)
	{
		return null;
	}
	
	/**
	 * This method does nothing. It has been overridden from HashMap return without 
	 * performing an operation. Instead, use the overloaded putAll method that takes
	 * a collection of AbstractRule.
	 * @param map
	 */
	public void putAll (Map<? extends Object, ? extends HashMap <Integer,AbstractRule>> map)
	{
		return;
	}
	
	/**
	 * This method does nothing. It has been overridden from HashMap to return null
	 * without performing an operation. Instead, use the overloaded remove method that takes
	 * an AbstractRule.
	 * @param key
	 * @return null
	 */
	public HashMap<Integer, AbstractRule> remove (Object key)
	{
		return null;
	}

	/**
	 * Puts a rule into the rule collection. If the specified rule is covered by any rules 
	 * currently in this collection, then the rule will be added to the collection of child 
	 * rules for that rule. If the specified rule covers any rules in this collection, those rules
	 * will be added to the child rules of the specified rule and the specified rule will 
	 * replace those rules in this collection. If the specified rule is equal to any rules 
	 * currently in this collection the rule will NOT be added.
	 * @param R The rule to add.
	 */
	public void put (AbstractRule R)
	{
		AbstractRule check = getCover(R.rCondition, R.rAction);
		if(check != null)
		{
			if(check.equals(R))
				return;
			R.resetChildren();
			Children.put(R);
			AbstractRule cr = Children.get(R.getCondition(), R.getAction());
			for(AbstractRule r : get(R.getAction().getID()).values())
			{
				if(r.covers(R))
				{
					if(r.getChildren() == null)
						r.initChildren();
					r.Children.put(cr.getCondition().hashCode(), cr);
				}
			}
		}
		else
		{
			if(R instanceof RefineableRule)
			{
				generateVariations((RefineableRule)R);
			}
			
			LinkedList <AbstractRule> delRules = new LinkedList<AbstractRule> ();
			HashMap<Integer,AbstractRule> rc = get(R.getAction().getID());
			if(rc != null)
			{
				for(AbstractRule c : rc.values())
				{
					if (R.getCondition().covers(c.getCondition()))
					{
						delRules.add(rc.get(c.getCondition().hashCode()));
					}
				}
				
				for(AbstractRule rr : delRules)
				{
					remove(rr.getCondition(), rr.getAction());
					rr.resetChildren();
					Children.put(rr);
					AbstractRule cr = Children.get(rr.getCondition(), rr.getAction());
					if(R.getChildren() == null)
						R.initChildren();
					R.Children.put(cr.getCondition().hashCode(), cr);
				}
				
				HashMap<Integer,AbstractRule> crc = Children.get(R.getAction().getID());
				if(crc != null)
				{
					for(AbstractRule r : crc.values())
					{
						if(R.covers(r))
						{
							if(R.Children == null)
								R.initChildren();
							R.Children.put(r.getCondition().hashCode(), r);
						}
					}
				}
			}
			
			removeChild(R);
			R.resetMatchStatistics();
			rc = get(R.getAction().getID());
			
			if(rc == null)
			{
				rc = new HashMap<Integer, AbstractRule> ();
				rc.put(R.getCondition().hashCode(), R);
				super.put(R.getAction().getID(), rc);
			}
			else
			{
				rc.put(R.getCondition().hashCode(), R);
			}
		}
	}
	
	/**
	 * Puts a collection of rules into the rule collection. If a rule in the specified 
	 * collection is equal to any rules currently in this collection, the rule will NOT 
	 * be added.
	 * @param R The rules to put in the collection.
	 */
	public void putAll (Collection <? extends AbstractRule> R)
	{
		for(AbstractRule r : R)
		{
			put(r);
		}
	}
	
	/**
	 * Removes the rule from this rule collection with the specified condition and action. This 
	 * method also removes all of the rule variations associated with the rule being removed as long
	 * as the variation is not also a variation for a different rule in the rule store.
	 * <p>
	 * If the rule collection does not contain a rule with the specified condition and action,
	 * this method returns null.
	 * @param cond The condition of the rule you want to remove.
	 * @param act The action of the rule you want to remove.
	 * @return The rule that was removed. If no rule was found in the collection, then null.
	 */
	public AbstractRule remove (GeneralizedConditionChunk cond, AbstractAction act)
	{
		HashMap<Integer,AbstractRule> rc = get(act.getID());
		if (rc == null)
			return null;
		AbstractRule r = rc.remove(cond.hashCode());
		if(r instanceof RefineableRule)
		{
			removeVariations((RefineableRule)r);
			Variations.remove(r.getCondition(), r.getAction());
			for(AbstractRule rr : rc.values())
			{
				if(rr instanceof RefineableRule)
				{
					((RefineableRule)rr).getVariations().remove(r.getCondition().hashCode());
				}
			}	
		}
		if(rc.size() == 0)
			super.remove(act.getID());
		return r;
	}
	
	/**
	 * Gets all of the rules in the rule collection.
	 * @return A collection containing the rules in the rule collection.
	 */
	public Collection <AbstractRule> getRules()
	{
		LinkedList <AbstractRule> rc = new LinkedList<AbstractRule> ();
		for(HashMap<Integer, AbstractRule> i : values())
		{
			rc.addAll(i.values());
		}
		return rc;
	}
	
	/**
	 * Removes the variations from the variations list for all variations of the 
	 * specified rule that are not also a variation for another rule in the rule collection.
	 * In addition, this method also removes the variations from the specified rule.
	 * @param R The rule whose variations you wish to remove.
	 */
	public void removeVariations (RefineableRule R)
	{
		for(AbstractRule vr : R.getVariations().values())
		{
			boolean removeVar = true;
			for(AbstractRule rr : get(R.getAction().getID()).values())
			{
				if(rr instanceof RefineableRule)
				{
					AbstractRule vrr = ((RefineableRule)rr).getVariations().get(vr.getCondition().hashCode());
					if(vrr != null)
					{
						removeVar = false;
						break;
					}
				}
			}
			if(removeVar)
				Variations.remove(vr.getCondition(), vr.getAction());
		}
		R.resetVariations();
	}
	
	/**
	 * Generates variations for the rule r and adds any variations that are not already
	 * in the variations list to that list. This method also adds the variations
	 * for the specified rule to that rule's internal variations list.
	 * @param R The rule from which to generate variations.
	 */
	private void generateVariations (RefineableRule R)
	{
		for(Dimension d : R.getCondition().values())
		{
			for(Value v : d.values())
			{
				RefineableRule r = new RefineableRule (R.getCondition().clone(),R.getAction());
				if(v.isFullyActivated())
				{
					r.getCondition().get(d.getID()).get(v.getID()).setActivation(v.MINIMUM_ACTIVATION_THRESHOLD);
				}
				else
					r.getCondition().get(d.getID()).get(v.getID()).setActivation(v.FULL_ACTIVATION_THRESHOLD);
				
				if(r.getCondition().get(d.getID()).getNumFullyActivatedVals() > 0 && !r.getCondition().checkMatchAll())
				{
					Variations.put(r);
					r = (RefineableRule)Variations.get(r.getCondition(), r.getAction());
					if(R.getVariations() == null)
						R.initVariations();
					R.getVariations().put(r.getCondition().hashCode(), r);
					r.setMatchAll(R.getMatchAll());
				}
			}
		}
	}
	
	/**
	 * Removes the specified rule from the child rule collection (if it is in the child rule collection).
	 * @param CR The rule to remove from the child rule collection.
	 */
	public void removeChild (AbstractRule CR)
	{
		Children.remove(CR.getCondition(), CR.getAction());
		HashMap <Integer, AbstractRule> crc = get(CR.getAction().getID());
		if(crc != null)
		{
			for(AbstractRule r : crc.values())
			{
				if(r.Children != null)
					r.Children.remove(CR.getCondition().hashCode());
			}
		}
	}
	
	/**
	 * Discounts the positive and negative match statistics for all rules in the rule 
	 * collection and all rule variations.
	 * @param Discount The discount factor to be applied to the match statistics of the rules.
	 */
	public void discountMatchStatistics(double Discount)
	{
		for(HashMap <Integer,AbstractRule> r : values())
		{
			for(AbstractRule i : r.values())
			{
				i.setPM(i.getPM() * Discount);
				i.setNM(i.getNM() * Discount);
			}
		}
		Variations.discountMatchStatistics(Discount);
		PM *= Discount;
		NM *= Discount;
	}
	
	/**
	 * Gets the positive match statistic.
	 * @return The positive match statistic.
	 */
	public double getPM ()
	{
		return PM;
	}
	
	/**
	 * Gets the negative match statistic.
	 * @return The negative match statistic.
	 */
	public double getNM ()
	{
		return NM;
	}
	
	/**
	 * Sets the positive match statistic. This method can be used to update the 
	 * positive match statistic if the user wishes to provide their own match criterion 
	 * function.
	 * @param pm The value to set as the positive match statistic.
	 */
	public void setPM (double pm)
	{
		PM = pm;
	}
	
	/**
	 * Sets the negative match statistic. This method can be used to update the 
	 * negative match statistic if the user wishes to provide their own match criterion 
	 * function.
	 * @param nm The value to set as the negative match statistic.
	 */
	public void setNM (double nm)
	{
		NM = nm;
	}
	
	/**
     * Gets the feedback. This method is only used if feedback is being provided.
     * @return The feedback.
     */
    public double getFeedback()
    {
    	return Feedback;
    }
    
    /**
     * Sets the feedback. This method should be called before the updateMatchStatistics method is called. 
     * This method is only used if feedback is being provided.
     * @param feedback The value of the feedback.
     */
    public void setFeedback (double feedback)
    {
    	Feedback = feedback;
    }
	
	/**
	 * Updates the positive or negative match statistics based on the feedback.
	 * <p>
	 * This update is usually performed after the feedback has been set.
	 * @param MatchCalculator The match calculator to use to determine positivity.
	 */
	public void updateMatchStatistics (AbstractMatchCalculator MatchCalculator)
	{
		if (MatchCalculator.isPositive(Feedback, POSITIVE_MATCH_THRESHOLD))
		{
			++PM;
		}
		else
			++NM;
	}
	
	/**
	 * Resets the match statistics.
	 */
	public void resetMatchStatistics ()
	{
		PM = 0;
		NM = 0;
	}
	
	/**
	 * Increments the positive match statistic. This method can be used to update the 
	 * positive match statistic if the user wishes to provide their own match criterion 
	 * function.
	 */
	public void incrementPM()
	{
		++PM;
	}
	
	/**
	 * Increments the negative match statistic. This method can be used to update the 
	 * negative match statistic if the user wishes to provide their own match criterion 
	 * function.
	 */
	public void incrementNM()
	{
		++NM;
	}
    
	/**
	 * This method does nothing as it is not used by the CLARION Library for this collection.
	 * @return False
	 */
	public boolean checkMatchCriterion() {
		return false;
	}
	
	/**
	 * Returns the number of rules in the rule collection.
	 * @return The number of rules.
	 */
	public int size ()
	{
		return getRules().size();
	}
	
	/**
	 * Gets the number of variations being tracked by this collection.
	 * @return The number of variations.
	 */
	public int numVariations ()
	{
		return Variations.size();
	}
	
	/**
	 * Gets the number of children being tracked by this collection.
	 * @return The number of children.
	 */
	public int numChildren ()
	{
		return Children.size();
	}
	
	/**
	 * Gets the rule variations from the rule collection. This method is used mainly for
	 * internal purposes and for reporting the rule variations (if desired).
	 * <p>
	 * This method should NOT be used to manipulate the rule variations from outside of the 
	 * CLARION library.
	 * @return A collection containing the rule variations in this rule collection.
	 */
	public Collection <AbstractRule> getVariations ()
	{
		return Variations.getRules();
	}
	
	/**
	 * Gets the child rules from the rule collection. This method is used mainly for
	 * internal purposes and for reporting the rule variations (if desired).
	 * <p>
	 * This method should NOT be used to manipulate the child rules from outside of the 
	 * CLARION library.
	 * @return A collection containing the child rule in this rule collection.
	 */
	public Collection <AbstractRule> getChildren ()
	{
		return Children.getRules();
	}
	
	public String toString()
	{
		String s = "";
		for(Iterator <AbstractRule> i = getRules().iterator(); i.hasNext();)
		{
			AbstractRule r = i.next();
			s += r.toString();
			if(i.hasNext())
				s+="\n";
		}
		return s;
	}
}
