package clarion.system;

import java.util.Collection;
import java.util.Map;

/**
 * This class implements a specialized working memory chunk collection within CLARION. 
 * It extends the AbstractChunkCollection class.
 * <p>
 * <b>Usage:</b>
 * <p>
 * This class acts as a container to hold specialized working memory chunks within the
 * working memory.
 * @version 6.0.4.5
 * @author Nick Wilson
 */
public class SpecializedWMChunkCollection extends AbstractChunkCollection<AbstractChunk> {
	private static final long serialVersionUID = 465679730864196253L;

	/**The hash for this object. When this item is cloned, so is this hash.*/
	private int hash;
	
	/**
	 * Initializes a specialized working memory collection.
	 */
	public SpecializedWMChunkCollection ()
	{
		super();
		hash = System.identityHashCode(this);
	}
	
	/**
	 * Initializes a specialized working memory chunk collection with the collection of 
	 * specialized working memory chunks specified.
	 * @param wmc The specialized working memory chunks for the collection.
	 */
	public SpecializedWMChunkCollection (Collection <? extends AbstractChunk> wmc)
	{
		super(wmc);
		hash = System.identityHashCode(this);
	}
	
	/**
	 * Initializes the specialized working memory chunk collection with the map of specialized
	 * working memory chunks.
	 * @param map The map of specialized working memory chunks for the collection.
	 */
	public SpecializedWMChunkCollection (Map <? extends Object, ? extends AbstractChunk> map)
	{
		super(map);
		hash = System.identityHashCode(this);
	}
	
	/**
	 * Returns the specialized working memory chunk collection as a dimension-value collection that 
	 * contains a dimension for each unique dimension within the chunks of the collection. For all
	 * chunks in the collection that contain the same dimension, the values within that
	 * dimension will all be placed under a single dimension within the dimension-value collection 
	 * that is returned.
	 * <p>
	 * This method is used to provide the specialized working memory chunks to the subsystems in a 
	 * format that can be used as input.
	 * @return The specialized working memory chunk collection as a dimension-value collection.
	 */
	public DimensionValueCollection toDimensionValueCollection ()
	{
		DimensionValueCollection a = new DimensionValueCollection ();
		for(AbstractChunk c : values())
		{
			for(Dimension d : c.values())
			{
				if(!a.containsKey(d.getID()))
					a.put(d.getID(), d);
				else
				{
					for(Value v : d.values())
						a.get(d.getID()).put(v.getID(), v);
				}
			}
		}
		return a;
	}
	
	/**
	 * Checks to see if the specified object is a specialized working memory chunk collection 
	 * and that all of the keys within the specified collection are specified within this 
	 * collection.
	 * @param WMCCollection The collection to compare to this collection.
	 * @return True if this collection contains all of the keys from the specified collection, 
	 * otherwise false.
	 */
	public boolean containsKeys (Object WMCCollection)
	{
		if (this == WMCCollection)
			return true;
		if (!(WMCCollection instanceof SpecializedWMChunkCollection))
			return false;
		for (Object k : ((SpecializedWMChunkCollection)WMCCollection).keySet())
		{
			if(!this.containsKey(k))
				return false;
		}
		return true;
	}

	/**
	 * Checks to see if the specified object is a specialized working memory chunk collection and that
	 * all of the chunks within the specified collection are equal to the chunks in this collection.
	 * @param WMCCollection The collection to compare to this collection.
	 * @return True if the two collections are equal, otherwise false.
	 */
	public boolean equals (Object WMCCollection)
	{
		if (this == WMCCollection)
			return true;
		if (!(WMCCollection instanceof SpecializedWMChunkCollection))
			return false;
		for (Map.Entry<? extends Object, ? extends AbstractChunk> e : ((SpecializedWMChunkCollection)WMCCollection).entrySet())
			if(!containsKey(e.getKey()) || !containsValue(e.getValue()))
				return false;
		return true;
	}
	
	public int hashCode()
	{
		return hash;
	}
	
	/**
	 * Clones the specialized working memory chunk collection (including all of it's chunks).
	 * @return A copy of the specialized working memory chunk collection.
	 */
	public SpecializedWMChunkCollection clone ()
	{
		SpecializedWMChunkCollection a = new SpecializedWMChunkCollection();
		for (AbstractChunk i : values())
		{
			AbstractChunk c = i.clone();
			a.put(c.getID(), c);
		}
		a.hash = hash;
		return a;
	}
}
