package clarion.system;

import java.util.Collection;
import java.util.Collections;

/**
 * This class implements the working memory within CLARION. It extends the AbstractIntermediateModule class.
 * Conceptually, this system exists between the ACS and the NACS. However,
 * implementationally it is logical to specify it as an intermediate module since it is used
 * by several subsystems. Therefore, in the CLARION Library, the working memory is contained within the
 * CLARION class.
 * <p>
 * <b>Usage:</b>
 * <p>
 * This class contains methods for adding and removing items from the working memory. The working memory is used
 * primarily by the ACS, but it can be used to facilitate interactions between the ACS and the NACS.
 * <p>
 * Working memory can contain specialized chunks that only exist within working memory.
 * Example of such a chunk include the "carry over" digit in addition or the "borrow" digit
 * in subtraction. Unlike a goal, the dimensions of a working memory chunk can have the same
 * IDs the dimensions in the sensory information space. This equates to "remembering" or
 * "visualizing" something within the sensory information space and recalling it for later use.
 * <p>
 * Currently this is simply a container for the working memory slots. Any manipulations on the chunks 
 * in working memory must be manually specified.
 * <p>
 * It is EXTREMELY important that you only manipulate the working memory using the add and remove
 * methods that have been overridden by this class.
 * <p>
 * This class contains both global (static) and local constants. The default is to use the local 
 * constants. If you want to change any of the global constants, you need to do so before any
 * instances of this class are initialized.
 * @version 6.0.4
 * @author Nick Wilson
 */
public class WorkingMemory extends AbstractIntermediateModule <AbstractChunk>{
	private static final long serialVersionUID = 3109928395636005779L;
	
	/**The number of slots in working memory.*/
	public static int GLOBAL_WORKING_MEMORY_CAPACITY = 7;
	/**The number of slots in working memory.*/
	public int WORKING_MEMORY_CAPACITY = GLOBAL_WORKING_MEMORY_CAPACITY;
	/**Specialized chunks specific only to working memory.*/
	private SpecializedWMChunkCollection SpecializedWMChunks;
	
	/**
	 * Initializes the working memory. During initialization this instance of the working
	 * memory will attach itself to the CLARION agent you specify.
	 * @param Agent The agent to which the working memory is being attached.
	 */
	public WorkingMemory (CLARION Agent) 
	{
		super(Agent);
	}
	
	/**
	 * Adds a chunk to working memory. If you are trying to add a chunk to working memory that has
	 * dimension-value pairs that do not exist in CLARION's internal input space or the working memory
	 * is full, this method will throw an exception.
	 * <p>
	 * If you want to add a new specialized working memory chunk to the list of specialized working
	 * memory chunks, call the addSpecializedWMChunk method.
	 * @param C The chunk to add.
	 * @return True only if the chunk was successfully added.
	 * @throws FullContainerException If working memory is full.
	 * @throws IllegalArgumentException If the input space does not contain all of the dimension-value
	 * pairs specified in the chunk.
	 */
	public boolean add (AbstractChunk C) throws FullContainerException, IllegalArgumentException
	{
		if(size() < WORKING_MEMORY_CAPACITY)
		{
			for(Dimension d : C.values())
			{
				Dimension a = InputSpace.get(d.getID());
				if(a == null)
					throw new IllegalArgumentException ("The specified chunk contains a dimension " +
							"that has not been specified within the agent's internal input space.");
				for(Value v : d.values())
				{
					if(!a.containsValue(v))
						throw new IllegalArgumentException ("The specified chunk contains a value " +
								"in dimension " + d.getID().toString() + " that has not been specified " +
								"within the agent's internal input space.");
				}
			}
			super.add(C);
		}
		else
			throw new FullContainerException();
		return true;
	}
	
	/**
	 * Adds a collection of chunks to working memory.
	 * @param C The chunks to add.
	 * @return True only if the chunks were successfully added.
	 */
	public boolean addAll (Collection <? extends AbstractChunk> C)
	{
		for(AbstractChunk c : C)
			add(c);
		return true;
	}
	
	/**
	 * Removes a chunk from working memory.
	 * @param Index The index of the chunk to remove.
	 * @return The removed chunk.
	 */
	public AbstractChunk remove (int Index)
	{
		AbstractChunk c = super.remove(Index);
		return c;
	}
	
	/**
	 * Removes the specified chunk object from working memory.
	 * @param C The chunk to remove.
	 * @return True if the chunk was successfully removed.
	 */
	public boolean remove (Object C)
	{
		super.remove(C);
		return true;
	}
	
	/**
	 * Gets the chunk from working memory with the specified ID.
	 * If none of the working memory chunks have the specified ID, this method 
	 * returns null.
	 * @param ID The ID of the chunk to get from the list of specialized working memory chunks.
	 * @return The chunk.
	 */
	public AbstractChunk get (Object ID)
	{
		for(AbstractChunk i : this)
		{
			if(i.getID().equals(ID))
				return i;
		}
		return null;
	}
	
	/**
	 * Gets the number of specialized working memory chunks.
	 * @return The number of specialized working memory chunks.
	 */
	public int getNumSpecializedWMChunks ()
	{
		return  SpecializedWMChunks.size();
	}
	
	/**
	 * Gets the chunk from the specialized working memory chunk list with the specified ID.
	 * If none of the specialized working memory chunks have the specified ID or the
	 * working memory does not contain any specialized working memory chunks, this method 
	 * returns null.
	 * <p>
	 * This does NOT return a chunk from working memory. To get a chunk from working memory, 
	 * call the get method.
	 * @param ID The ID of the chunk to get from the list of specialized working memory chunks.
	 * @return The chunk.
	 */
	public AbstractChunk getSpecializedWMChunk (Object ID)
	{
		if(SpecializedWMChunks != null)
		{
			for(AbstractChunk i : SpecializedWMChunks.values())
			{
				if(i.getID().equals(ID))
					return i;
			}
		}
		return null;
	}
	
	/**
	 * Gets the chunk from the specialized working memory chunk list at the specified Index. If the
	 * working memory does not contain any specialized working memory chunks, this method returns null.
	 * <p>
	 * This does NOT return a chunk from working memory. To get a chunk from working memory, 
	 * call the get method.
	 * @param Index The index of the chunk to get from the list of specialized working memory chunks.
	 * @return The chunk.
	 */
	public AbstractChunk getSpecializedWMChunk (int Index)
	{
		if(SpecializedWMChunks != null)
			return SpecializedWMChunks.get(Index);
		return null;
	}
	
	/**
	 * Gets a collection of the specialized working memory chunks. The collection returned is 
	 * unmodifiable and is meant for reporting the internal state only. If the working memory does 
	 * not contain any specialized working memory chunks, this method returns null.
	 * @return An unmodifiable collection of specialized working memory chunks.
	 */
	public Collection <AbstractChunk> getSpecializedWMChunks ()
	{
		if(SpecializedWMChunks != null)
			return Collections.unmodifiableCollection(SpecializedWMChunks.values());
		return null;
	}
	
	/**
	 * Adds a working memory chunk to the list of specialized working memory chunks. If the 
	 * chunk is already in the list of specialized working memory chunks, it will not be added.
	 * <p>
	 * Remember that the input layer of an implicit module or the condition of a rule in the ACS 
	 * must contain nodes for all dimension-value pairs contained within any of the
	 * specialized working memory chunks you wish to have as input into the implicit module or 
	 * rule condition.
	 * <p>
	 * This does NOT add a chunk to working memory. To add a chunk to working memory, 
	 * call the add method.
	 * @param C The specialized working memory chunk to add.
	 * @return True if the chunk was successfully added.
	 */
	public boolean addSpecializedWMChunk (AbstractChunk C)
	{
		if(SpecializedWMChunks == null)
		{
			SpecializedWMChunks = new SpecializedWMChunkCollection ();
		}
		
		if(!SpecializedWMChunks.containsKey(C.getID()))
		{
			AbstractChunk c = C.clone();
			updateInputSpace(c.values());
			SpecializedWMChunks.put(c.getID(), c);
			return true;
		}
		return false;
	}
	
	/**
	 * Adds a collection of specialized working memory chunks to the list of
	 * specialized working memory chunks. If a chunk from the collection 
	 * is already in the list of specialized working memory chunks, it will not be added.
	 * <p>
	 * Remember that the input layer of an implicit module or the condition of a rule in the ACS 
	 * must contain nodes for all dimension-value pairs contained within any of the
	 * specialized working memory chunks you wish to have as input into the implicit module 
	 * or rule condition.
	 * <p>
	 * This does NOT add a collection of chunks to working memory. To add a collection of 
	 * chunks to working memory, call the addAll method.
	 * @param Chunks The collection of specialized working memory chunks to add.
	 * @return True if any of the specialized working memory chunks were successfully added.
	 */
	public boolean addSpecializedWMChunks (Collection <AbstractChunk> Chunks)
	{
		boolean rettrue = false;
		for(AbstractChunk c : Chunks)
		{
			if(addSpecializedWMChunk(c));
				rettrue = true;
		}
		return rettrue;
	}
	
	/**
	 * Attaches the working memory to the specified CLARION agent.
	 * @param Agent The agent to wish this working memory will be attached.
	 */
	protected void attachSelfToAgent (CLARION Agent)
	{
		Agent.attachWorkingMemory(this);
	}
}
