package clarion.tools;

import java.util.HashMap;
import clarion.system.Drive;
import clarion.system.DriveCollection;

/**
 * This class implements a personality model loosely based on the primary drives posited in CLARION.
 * This class contains only enumerators and therefore does not need to be instantiated.
 * <p>
 * <b>Usage:</b>
 * <p>
 * This class is primarily used to setup drives within the motivational subsystem (MS) in CLARION. It
 * provides enumerators that can assist in speeding up the process of initializing the MS.
 * <p>
 * For information regarding the theory behind this implementation of the personality model, please see the 
 * addendum to the CLARION tutorial on the MS (Sun, 2008).
 * @version 6.0.5
 * @author Nick Wilson
 */
public class Personality {
	/**The primary drives proposed by CLARION.*/
	public enum PrimaryDrives {FOOD, WATER, REST, AVOIDING_PHYSICAL_DANGER, REPRODUCTION,
		AVOIDING_THE_UNPLEASANT, AFFILIATION_AND_BELONGINGNESS, RECOGNITION_AND_ACHEIVEMENT,
		DOMINANCE_AND_POWER, AUTONOMY, DEFERENCE, SIMILANCE, FAIRNESS, HONOR, NURTURANCE, 
		CONSERVATION, CURIOSITY};
	/**Default personality types that can be used for initializing the MS.*/
	public enum PersonalityTypes {
		GENERIC (.1,.1,.05,.2,.1,.2,.6,.2,.2,.5,.3,.7,.1,.5,.4,.3,.4);
		
		/**The drive deficits used to define the personality.*/
		public final DriveCollection Drives;
		/**
		 * Initializes a personality type from the list of deficits specified. The order of the 
		 * deficits must be the same as the order from the PrimaryDrives enumerator.
		 * @param deficits The deficits to set for the drives.
		 */
		private PersonalityTypes (double... deficits)
		{
			Drives = new DriveCollection();
			for(int i = 0; i < deficits.length; i++)
			{
				Drives.put(PrimaryDrives.values()[i], new Drive (PrimaryDrives.values()[i], deficits[i]));
			}
		}
	};
	
	/**For the PrimaryDrives enumerator, this specifies to which behavioral system they belong.
	 * Using this enum you can find out which of the PrimaryDrives belong to:<br>
	 * <ul>
	 * <li>The BAS only</li>
	 * <li>The BIS only</li>
	 * <li>Both systems</li>
	 * </ul>
	 * The members of the systems are stored in the PrimaryDrives HashMap accessed by typing, for example:
	 * <p>
	 * <i>Personality.BehavioralSystemMembers.BAS_ONLY.PrimaryDrives</i>
	 * <p>
	 * The keys for the PrimaryDrives HashMap is the ordinal value of the primary drive in the PrimaryDrives enumerator
	 * located in the Personality class.
	 */
	public enum BehavioralSystemMembers {
		BAS_ONLY (Personality.PrimaryDrives.FOOD, Personality.PrimaryDrives.WATER, Personality.PrimaryDrives.REPRODUCTION,
				Personality.PrimaryDrives.RECOGNITION_AND_ACHEIVEMENT, Personality.PrimaryDrives.DOMINANCE_AND_POWER,
				Personality.PrimaryDrives.NURTURANCE, Personality.PrimaryDrives.CURIOSITY),
		BIS_ONLY (Personality.PrimaryDrives.REST, Personality.PrimaryDrives.AVOIDING_PHYSICAL_DANGER, 
				Personality.PrimaryDrives.AVOIDING_THE_UNPLEASANT, Personality.PrimaryDrives.HONOR,
				Personality.PrimaryDrives.CONSERVATION),
		BOTH_SYSTEMS (Personality.PrimaryDrives.AFFILIATION_AND_BELONGINGNESS, Personality.PrimaryDrives.AUTONOMY,
				Personality.PrimaryDrives.DEFERENCE, Personality.PrimaryDrives.SIMILANCE, Personality.PrimaryDrives.FAIRNESS);
		
		/**The collection of primary drives contained within the behavioral system.
		 * The key is the ordinal value of the primary drive in the PrimaryDrives enumerator
		 * located in the Personality class*/
		public final HashMap <Integer, Personality.PrimaryDrives> PrimaryDrives = 
			new HashMap<Integer, Personality.PrimaryDrives> ();
		
		/**
		 * Initializes the enumerated types for this enumerator.
		 * @param d The primary drives associated with the behavioral system identified by the enumerated type.
		 */
		private BehavioralSystemMembers (Personality.PrimaryDrives... d){
			for(Personality.PrimaryDrives i : d)
				PrimaryDrives.put(i.ordinal(), i);
		}
	};
}
